#include <stdio.h>
#include "arm.h"
#define uchar unsigned char

/*  uses JDG's arm_v3

notes:

  random changed to rand to allow linking
  N.B. check rand as may not produce 32-bit no.s


needs:

  1. long multiply  - draughted BUT NOT TESTED
  2. btb updating for thumb
  3. more thumb testing

lsl routine added but not fully tested (28/10/98)
Long multiplication written (27/1/00)
Architecture V5 added (2/2/00)

To do:
   check long muls
         Flag checking (immediates ?!)
      Validation
   interrupt enable behaviour on exceptions (etc.)
*/
/* NB "int" is assumed to be at least 32-bit */

/* Stripped exectution code (9/8/00) */

/* ARM disassemble */
		
const unsigned int nf_mask        = 0X80000000;
const unsigned int zf_mask        = 0X40000000;
const unsigned int cf_mask        = 0X20000000;
const unsigned int vf_mask        = 0X10000000;
const unsigned int if_mask        = 0X00000080;
const unsigned int ff_mask        = 0X00000040;
const unsigned int mode_mask      = 0X0000001F;
const unsigned int tf_mask        = 0X00000020;                  /* THUMB bit */

const unsigned int bit_31         = 0X80000000;
const unsigned int bit_0          = 0X00000001;

const unsigned int imm_mask       = 0X02000000;      /* orginal word versions */
const unsigned int imm_hw_mask    = 0X00400000;         /* half word versions */
const unsigned int data_op_mask   = 0X01E00000;          /* ALU function code */
const unsigned int data_ext_mask  = 0X01900000;   /* To sort out CMP from MRS */
const unsigned int arith_ext      = 0X01000000; /* Poss. arithmetic extension */
const unsigned int s_mask         = 0X00100000;
const unsigned int rn_mask        = 0X000F0000;
const unsigned int rd_mask        = 0X0000F000;
const unsigned int rs_mask        = 0X00000F00;
const unsigned int rm_mask        = 0X0000000F;
const unsigned int op2_mask       = 0X00000FFF;
const unsigned int hw_mask        = 0X00000020;
const unsigned int sign_mask      = 0X00000040;

const unsigned int mul_mask       = 0X0FC000F0;
const unsigned int long_mul_mask  = 0X0F8000F0;
const unsigned int mul_op         = 0X00000090;
const unsigned int long_mul_op    = 0X00800090;
const unsigned int mul_acc_bit    = 0X00200000;
const unsigned int mul_sign_bit   = 0X00400000;
const unsigned int mul_long_bit   = 0X00800000;

const unsigned int sbhw_mask      = 0X0E000FF0;

const unsigned int swp_mask       = 0X0FB00FF0;
const unsigned int swp_op         = 0X01000090;

const unsigned int pre_mask       = 0X01000000;
const unsigned int up_mask        = 0X00800000;
const unsigned int byte_mask      = 0X00400000;
const unsigned int write_back_mask= 0X00200000;
const unsigned int load_mask      = 0X00100000;
const unsigned int byte_sign      = 0X00000080;
const unsigned int hw_sign        = 0X00008000;

const unsigned int user_mask      = 0X00400000;

const unsigned int link_mask      = 0X01000000;
const unsigned int branch_field   = 0X00FFFFFF;
const unsigned int branch_sign    = 0X00800000;

const unsigned int undef_mask     = 0X0E000010;
const unsigned int undef_code     = 0X06000010;
const int TRUE = -1;
const int FALSE = 0;


/*----------------------------------------------------------------------------*/
/* Entry point */


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
int is_it_sbhw(unsigned int op_code)
{
if (((op_code & 0X0E000090) == 0X00000090)
 && ((op_code & 0X00000060) != 0X00000000)                  /* No multiplies */
 && ((op_code & 0X00100040) != 0X00000040))              /* No signed stores */
  {
  if (((op_code & 0X00400000) != 0) || ((op_code & 0X00000F00) == 0))
    return TRUE;
  else
    return FALSE;
  }
else
  return FALSE;
}
/*----------------------------------------------------------------------------*/
/* As the compiler can't manage it ...                                        */

int lsl(int value, int distance, int *cf)         /* cf is -internal- Boolean */
{
int result;

if (distance != 0)
  {
  if (distance < 32)
    {
    result = value << distance;
    *cf = (((value << (distance-1)) & bit_31) != 0);
   }
  else
    {
    result = 0X00000000;
    if (distance == 32) *cf = ((value & bit_0) != 0);
    else                *cf = (0 != 0);             /* internal "false" value */
    }
  }
else
  result = value;

return result;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int lsr(unsigned int value, int distance, int *cf)/* cf is -internal- Boolean */
{
unsigned int result, mask;

if (distance != 0)
  {
  if (distance < 32)
    {
    if (distance != 0) mask = ~(0XFFFFFFFF << (32-distance));
    else               mask =   0XFFFFFFFF;
    result = (value >> distance) & mask;             /* Enforce logical shift */
    *cf = (((value >> (distance-1)) & bit_0) != 0);
    }
  else
    {                             /* Make a special case because C is so crap */
    result = 0X00000000;
    if (distance == 32) *cf = ((value & bit_31) != 0);
    else                *cf = (0 != 0);             /* internal "false" value */
    }
  }
else
  result = value;

return result;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int asr(int value, int distance, int *cf)         /* cf is -internal- Boolean */
{
int result;

if (distance != 0)
  {
  if (distance < 32)
    {
    result = value >> distance;
    if (((value & bit_31) != 0) && (distance != 0))
      result = result | (0XFFFFFFFF << (32-distance));
                                  /* Sign extend - I don't trust the compiler */
    *cf = (((value >> (distance-1)) & bit_0) != 0);
    }
  else
    {                             /* Make a special case because C is so crap */
    *cf = ((value & bit_31) != 0);
    if ((value & bit_31) == 0) result = 0X00000000;
    else                       result = 0XFFFFFFFF;
    }
  }
else
  result = value;

return result;
}



int ror(unsigned int value, int distance, int *cf)/* cf is -internal- Boolean */
{
int result;

if (distance != 0)
  {
  distance = distance & 0X1F;
  result = lsr(value, distance, cf)
         | lsl(value, 32 - distance, cf);            /* cf acts as dummy here */
  *cf = (((value >> (distance-1)) & bit_0) != 0);
  }
else
  result = value;

return result;
}


void print_reg_list(int i)                      /* Assumes "i" already masked */
{
int state, k;

state = 0;                                            /* little state machine */
k     = 0;

while (i != 0)
  {
  switch (state)
    {
    case 0:
        if ((i & 1) != 0) {dis_string_ptr += sprintf(dis_string_ptr,"R%d", k); state = 1;}
        break;
    case 1:
        if ((i & 1) != 0) {dis_string_ptr += sprintf(dis_string_ptr,"-"); state = 2;}
        else              {dis_string_ptr += sprintf(dis_string_ptr,","); state = 0;}
        break;
    case 2:
        if ((i & 1) == 0)
          {
          dis_string_ptr += sprintf(dis_string_ptr,"R%d", k - 1); if (i != 0) dis_string_ptr += sprintf(dis_string_ptr,",");
          state = 0;
          }
        break;
    }
  i = i >> 1;
  k = k + 1;
  }                                                           /* End of WHILE */

if (state==2) dis_string_ptr += sprintf(dis_string_ptr,"R%d", k - 1);

return;
}

void arm_disassemble(unsigned int address, unsigned int op_code)
{

unsigned int cc;

cc = (op_code>>28) & 0X0000000F;
switch ((op_code>>25) & 0X00000007)
  {
  case 0X0: dis_data_op(address, op_code, cc);break;
  case 0X1: dis_data_op(address, op_code, cc);break;
  case 0X2: dis_transfer(address, op_code, cc);break;
  case 0X3: dis_transfer(address, op_code, cc);break;
  case 0X4: dis_multiple(address, op_code, cc);break;
  case 0X5: dis_branch(address, op_code, cc);break;
  case 0X6: dis_coprocessor(address, op_code, cc);break;
  case 0X7: dis_system(address, op_code, cc);break;
  }


return;
}

/*----------------------------------------------------------------------------*/

void dis_data_op(unsigned int address, unsigned int op_code, int cc)
{
int operation;
/*int shift_carry;            CB unused crying*/

if (((op_code & mul_mask) == mul_op)                  /* Short multiplication */
 || ((op_code & long_mul_mask) == long_mul_op))   /*  and long multiplication */
  dis_mul(address, op_code, cc);
else
  if (is_it_sbhw(op_code) == TRUE) dis_sbhw(address, op_code, cc);
  else
    if ((op_code & swp_mask) == swp_op) dis_swp(address, op_code, cc);
    else
      {

      operation = (op_code & data_op_mask)>>21;

      if ((operation >= 0X8) && (operation <= 0XB) && ((op_code & s_mask) == 0))
        {                                              /* PSR transfers OR BX */
        if ((op_code & 0X0FBF0FFF) == 0X010F0000)                      /* MRS */
          dis_mrs(address, op_code, cc);
        else                                                           /* MSR */
          if (((op_code & 0X0DB6F000) == 0X0120F000)
           && ((op_code & 0X02000010) != 0X00000010))
            dis_msr(address, op_code, cc);
          else                                                      /* BX/BLX */
            if ((op_code & 0X012FFFD0) == 0X012FFF10)
              dis_bx(address, op_code, cc);
            else                                  /* Notice no condition code */
              if ((op_code & 0XFFF000F0) == 0XE1200070) dis_string_ptr += sprintf(dis_string_ptr,"Breakpoint");
              else
                if ((op_code & 0X0FFF0FF0) == 0X016F0F10)              /* CLZ */
                  dis_dat(address, op_code, cc, operation);
                else dis_string_ptr += sprintf(dis_string_ptr,"UNDEFINED");
        }
      else                                 /* data processing immediate shift */
        dis_dat(address, op_code, cc, operation);
      }

return;
}

/*----------------------------------------------------------------------------*/

void dis_mrs(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2 + 1;
dis_string_ptr += sprintf(dis_string_ptr,"MRS");

dis_cc(cc, &spaces);
dis_tab(spaces);

dis_string_ptr += sprintf(dis_string_ptr,"R%d, ",(op_code & rd_mask)>>12);
if ((op_code & 0X00400000) == 0) dis_string_ptr += sprintf(dis_string_ptr,"CPSR");
else                             dis_string_ptr += sprintf(dis_string_ptr,"SPSR");

return;
}

/*----------------------------------------------------------------------------*/

void dis_msr(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2 + 1;
dis_string_ptr += sprintf(dis_string_ptr,"MSR");

dis_cc(cc, &spaces);
dis_tab(spaces);

if ((op_code & 0X00400000) == 0) dis_string_ptr += sprintf(dis_string_ptr,"CPSR");
else                             dis_string_ptr += sprintf(dis_string_ptr,"SPSR");

switch (op_code & 0X00090000)
  {
  case 0X00000000: dis_string_ptr += sprintf(dis_string_ptr,"_none?"); break;
  case 0X00010000: dis_string_ptr += sprintf(dis_string_ptr,"_mode");  break;
  case 0X00080000: dis_string_ptr += sprintf(dis_string_ptr,"_flg");   break;
  case 0X00090000: break;
  }

dis_string_ptr += sprintf(dis_string_ptr,", ");

if ((op_code & imm_mask) == 0)                 /* Test applies for both cases */
  dis_string_ptr += sprintf(dis_string_ptr,"R%d",(op_code & rm_mask));
else
  dis_immediate(op_code & 0X00000FFF);

return;
}

/*----------------------------------------------------------------------------*/

void dis_bx(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

if ((op_code & 0X00000020) == 0) {dis_string_ptr += sprintf(dis_string_ptr,"BX");  spaces = 4;}
else                             {dis_string_ptr += sprintf(dis_string_ptr,"BLX"); spaces = 3;}

dis_cc(cc, &spaces);
dis_tab(spaces);

dis_string_ptr += sprintf(dis_string_ptr,"R%d ", (op_code & rm_mask));
}

/*----------------------------------------------------------------------------*/

void dis_sbhw(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2;

if ((op_code & 0X00100000) != 0) dis_string_ptr += sprintf(dis_string_ptr,"LDR"); else dis_string_ptr += sprintf(dis_string_ptr,"STR");

dis_cc(cc, &spaces);

switch (op_code & 0X00000060)
  {
  case 0X20: dis_string_ptr += sprintf(dis_string_ptr,"H");  break;
  case 0X40: dis_string_ptr += sprintf(dis_string_ptr,"SB"); break;
  case 0X60: dis_string_ptr += sprintf(dis_string_ptr,"SH"); break;
  }

dis_tab(spaces);

dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rd_mask) >> 12);
dis_string_ptr += sprintf(dis_string_ptr,"[R%d",  (op_code & rn_mask) >> 16);

if ((op_code & pre_mask) == 0)                                  /* Post index */
  {
  dis_string_ptr += sprintf(dis_string_ptr,"], ");
  dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                      op_code & imm_hw_mask, TRUE);
  }
else                                                             /* Pre index */
  {
  dis_string_ptr += sprintf(dis_string_ptr,", ");
  dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                      op_code & imm_hw_mask, TRUE);
  dis_string_ptr += sprintf(dis_string_ptr,"]");
  if ((op_code & write_back_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"!");   /* Optional write back */
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_dat(unsigned int address, unsigned int op_code, int cc, int operation)
{
int spaces;

spaces = 2;                                                    /* Format info */

switch (operation)
  {
  case 0X0:dis_string_ptr += sprintf(dis_string_ptr,"AND"); break;
  case 0X1:dis_string_ptr += sprintf(dis_string_ptr,"EOR"); break;
  case 0X2:dis_string_ptr += sprintf(dis_string_ptr,"SUB"); break;
  case 0X3:dis_string_ptr += sprintf(dis_string_ptr,"RSB"); break;
  case 0X4:dis_string_ptr += sprintf(dis_string_ptr,"ADD"); break;
  case 0X5:dis_string_ptr += sprintf(dis_string_ptr,"ADC"); break;
  case 0X6:dis_string_ptr += sprintf(dis_string_ptr,"SBC"); break;
  case 0X7:dis_string_ptr += sprintf(dis_string_ptr,"RSC"); break;
  case 0X8:dis_string_ptr += sprintf(dis_string_ptr,"TST"); break;
  case 0X9:dis_string_ptr += sprintf(dis_string_ptr,"TEQ"); break;
  case 0XA:dis_string_ptr += sprintf(dis_string_ptr,"CMP"); break;
  case 0XB:
    if ((op_code & s_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"CMN");
    else                         dis_string_ptr += sprintf(dis_string_ptr,"CLZ");
    break;
  case 0XC:dis_string_ptr += sprintf(dis_string_ptr,"ORR"); break;
  case 0XD:dis_string_ptr += sprintf(dis_string_ptr,"MOV"); break;
  case 0XE:dis_string_ptr += sprintf(dis_string_ptr,"BIC"); break;
  case 0XF:dis_string_ptr += sprintf(dis_string_ptr,"MVN"); break;
  }

dis_cc(cc, &spaces);

switch (operation)
  {
  case 0X0: case 0X1: case 0X2: case 0X3:
  case 0X4: case 0X5: case 0X6: case 0X7:
  case 0XC: case 0XD: case 0XE: case 0XF:
    if ((op_code & s_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"S"); else spaces++;
    break;

  case 0X8: case 0X9: case 0XA: case 0XB:
    if ((op_code & rd_mask) == 0XF) dis_string_ptr += sprintf(dis_string_ptr,"P"); else spaces++;
    break;
  }

dis_tab(spaces);

if ((op_code & 0X0FFF0FF0) == 0X016F0F10)
  {
  operation = 13;                              /* CLZ disassembles like a MOV */
  op_code = (op_code & 0XF000F00F) | 0X01A00000;           /*  with no shifts */
  };

switch (operation)
  {
  case 0X0: case 0X1: case 0X2: case 0X3:
  case 0X4: case 0X5: case 0X6: case 0X7:
  case 0XC:           case 0XE:
    dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rd_mask)>>12);
    dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rn_mask)>>16);
  break;

  case 0X8: case 0X9: case 0XA: case 0XB:
    dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rn_mask)>>16);
  break;

  case 0XD:                     case 0XF:
    dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rd_mask)>>12);
   break;

  }

if ((op_code & imm_mask) == 0) dis_shift(op_code);
else                           dis_immediate(op_code & 0X00000FFF);

return;
}

/*----------------------------------------------------------------------------*/

void dis_shift(unsigned int op_code)
{
int reg, type, distance, rs;

reg = op_code & 0X00F;                                     /* Register number */
type = (op_code & 0X060)>>5;                                 /* Type of shift */
if ((op_code & 0X010) == 0)
  {                                                        /* Immediate value */
  distance = (op_code & 0XF80)>>7;
  if (distance == 0)
    {
    if (type == 3) type = 4;                                           /* RRX */
    else if (type != 0) distance = 32;
    }

  switch (type)
    {
    case 0X0:
      if (distance == 0) dis_string_ptr += sprintf(dis_string_ptr,"R%d", reg);
      else               dis_string_ptr += sprintf(dis_string_ptr,"R%d LSL #%d", reg, distance);
    break;
    case 0X1:dis_string_ptr += sprintf(dis_string_ptr,"R%d LSR #%d", reg, distance); break;
    case 0X2:dis_string_ptr += sprintf(dis_string_ptr,"R%d ASR #%d", reg, distance); break;
    case 0X3:dis_string_ptr += sprintf(dis_string_ptr,"R%d ROR #%d", reg, distance); break;
    case 0X4:dis_string_ptr += sprintf(dis_string_ptr,"R%d RRX #1",  reg);           break;
    }
  }
else
  {
  rs = (op_code >> 8) & 0XF;
  switch (type)
    {
    case 0X0:dis_string_ptr += sprintf(dis_string_ptr,"R%d LSL R%d", reg, rs); break;
    case 0X1:dis_string_ptr += sprintf(dis_string_ptr,"R%d LSR R%d", reg, rs); break;
    case 0X2:dis_string_ptr += sprintf(dis_string_ptr,"R%d ASR R%d", reg, rs); break;
    case 0X3:dis_string_ptr += sprintf(dis_string_ptr,"R%d ROR R%d", reg, rs); break;
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_immediate(unsigned int op2)
{
unsigned int dummy;

dis_string_ptr += sprintf(dis_string_ptr,"#%X", ror(op2 & 0X0FF, (op2 & 0XF00)>>7, &dummy));
return;
}

/*----------------------------------------------------------------------------*/

void dis_mul(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

if ((op_code & mul_long_bit) != 0)
  {
  spaces = 0;
  if ((op_code & mul_sign_bit) != 0) dis_string_ptr += sprintf(dis_string_ptr,"S");
  else                               dis_string_ptr += sprintf(dis_string_ptr,"U");

  if ((op_code & mul_acc_bit) != 0) dis_string_ptr += sprintf(dis_string_ptr,"MLAL");
  else                              dis_string_ptr += sprintf(dis_string_ptr,"MULL");

  dis_cc(cc, &spaces);

  if ((op_code & s_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"S"); else spaces++;

  dis_tab(spaces);

  dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rd_mask)>>12);
  dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rn_mask)>>16);
  dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rm_mask));
  dis_string_ptr += sprintf(dis_string_ptr,"R%d",   (op_code & rs_mask)>>8);
  }
else
  {
  spaces = 2;
  if ((op_code & mul_acc_bit) != 0) dis_string_ptr += sprintf(dis_string_ptr,"MLA");
  else                              dis_string_ptr += sprintf(dis_string_ptr,"MUL");

  dis_cc(cc, &spaces);

  if ((op_code & s_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"S"); else spaces++;

  dis_tab(spaces);

  dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rn_mask)>>16);
  dis_string_ptr += sprintf(dis_string_ptr,"R%d, ", (op_code & rm_mask));
  dis_string_ptr += sprintf(dis_string_ptr,"R%d",   (op_code & rs_mask)>>8);
  if ((op_code & mul_acc_bit) != 0) dis_string_ptr += sprintf(dis_string_ptr,", R%d", (op_code & rd_mask)>>12);
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_swp(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2;

dis_string_ptr += sprintf(dis_string_ptr,"SWP");
dis_cc(cc, &spaces);

if ((op_code & byte_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"B");
else                            spaces++;
dis_tab(spaces);

dis_string_ptr += sprintf(dis_string_ptr,"R%d, R%d, [R%d]", (op_code & rd_mask)>>12,
                          (op_code & rm_mask),
                          (op_code & rn_mask)>>16);
return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2;

if ((op_code & undef_mask) == undef_code)
  dis_string_ptr += sprintf(dis_string_ptr,"Undefined");
else
  {
  if ((op_code & load_mask) == 0) dis_string_ptr += sprintf(dis_string_ptr,"STR");
  else                            dis_string_ptr += sprintf(dis_string_ptr,"LDR");

  dis_cc(cc, &spaces);

  if ((op_code & byte_mask) == 0) spaces++; else dis_string_ptr += sprintf(dis_string_ptr,"B");

  if (((op_code & pre_mask) == 0) && ((op_code & write_back_mask) != 0))
    {dis_string_ptr += sprintf(dis_string_ptr,"T"); spaces--;}

  dis_tab(spaces);

  dis_string_ptr += sprintf(dis_string_ptr,"R%d, ",(op_code & rd_mask)>>12);

/* Trap out PC relative as "absolute" */
  if ((op_code & 0X032F0000) == 0X010F0000)
    {
    int offset;
    offset = op_code & 0X00000FFF;
    if ((op_code & 0X00800000) == 0) offset = -offset;
    dis_string_ptr += sprintf(dis_string_ptr,"%08X", address + offset + 8);
    }
  else
    {
    dis_string_ptr += sprintf(dis_string_ptr,"[R%d",(op_code & rn_mask)>>16);
    if ((op_code & pre_mask) == 0)
      {
      dis_string_ptr += sprintf(dis_string_ptr,"], ");
      dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                         ~op_code & imm_mask, FALSE); /* NB Imm. bit inverted */
      }
    else
      {
      dis_string_ptr += sprintf(dis_string_ptr,", ");
      dis_transfer_offset(op_code & 0X00000FFF, op_code & up_mask,
                         ~op_code & imm_mask, FALSE); /* NB Imm. bit inverted */
      dis_string_ptr += sprintf(dis_string_ptr,"]");
      if ((op_code & write_back_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"!");
      }
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer_offset(int op2, int add, int imm, int sbhw)
{
/* sbhw bit 22 = 1 imm, = 0 reg
        others bit 25 = 0 imm, 1 = reg */

if ((imm == 0) || (sbhw == TRUE) || ((op2 & 0XFFF) != 0))/* Skip if no offset */
  {
  if (add == 0) dis_string_ptr += sprintf(dis_string_ptr,"-"); else dis_string_ptr += sprintf(dis_string_ptr,"+");

  if (sbhw == 0)                                      /* word - unsigned byte */
    {
    if (imm == 0) dis_shift(op2);                                  /* bit(25) */
    else dis_string_ptr += sprintf(dis_string_ptr,"#%X", (op2 & 0XFFF));
    }
  else
    {
    if (imm != 0)                                    /* half word signed byte */
      dis_string_ptr += sprintf(dis_string_ptr,"#%02X", ((op2 & 0XF00) >> 4) | (op2 & 0X00F));
    else
      dis_shift(op2);
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_multiple(unsigned int address, unsigned int op_code, int cc)
{
int spaces;

spaces = 2 - 1;

if ((op_code & load_mask) == 0) dis_string_ptr += sprintf(dis_string_ptr,"STM");
else                            dis_string_ptr += sprintf(dis_string_ptr,"LDM");

dis_cc(cc, &spaces);

switch ((op_code & 0X01800000)>>23)
  {
  case 0: dis_string_ptr += sprintf(dis_string_ptr,"DA"); break;
  case 1: dis_string_ptr += sprintf(dis_string_ptr,"IA"); break;
  case 2: dis_string_ptr += sprintf(dis_string_ptr,"DB"); break;
  case 3: dis_string_ptr += sprintf(dis_string_ptr,"IB"); break;
  }

dis_tab(spaces);

dis_string_ptr += sprintf(dis_string_ptr,"R%d",(op_code & rn_mask)>>16);
if ((op_code & write_back_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"!");

dis_string_ptr += sprintf(dis_string_ptr,", {"); print_reg_list(op_code & 0X0000FFFF); dis_string_ptr += sprintf(dis_string_ptr,"}");

if ((op_code & user_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"^");

return;
}

/*----------------------------------------------------------------------------*/

void dis_branch(unsigned int address, unsigned int op_code, int cc)
{
int offset;
int spaces;

offset = (op_code & branch_field) << 2;

if ((op_code & 0XF0000000) == 0XF0000000)                    /* Check for BLX */
  {
  dis_string_ptr += sprintf(dis_string_ptr,"BLX     ");
  offset = offset | ((op_code >> 23) & 2);                    /* Get bit 1 in */
  }
else
  {
  spaces = 4;
  dis_string_ptr += sprintf(dis_string_ptr,"B");
  if ((op_code & link_mask) != 0) dis_string_ptr += sprintf(dis_string_ptr,"L"); else spaces++;

  dis_cc(cc, &spaces);
  dis_tab(spaces);
  }

if ((op_code & branch_sign) != 0)
  offset = offset | (~(branch_field<<2) & 0XFFFFFFFC);         /* sign extend */

dis_string_ptr += sprintf(dis_string_ptr,"%08X", offset + address + 8);

return;
}

/*----------------------------------------------------------------------------*/

void dis_coprocessor(unsigned int address, unsigned int op_code, int cc)
{
dis_string_ptr += sprintf(dis_string_ptr,"coprocessor");
}

/*----------------------------------------------------------------------------*/

void dis_system(unsigned int address, unsigned int op_code, int cc)
{
if ((op_code & 0X0F000000) != 0X0F000000)
  dis_string_ptr += sprintf(dis_string_ptr,"Illegal");
else
  {
  int spaces;

  spaces = 2+1;
  dis_string_ptr += sprintf(dis_string_ptr,"SWI");
  dis_cc(cc, &spaces);
  dis_tab(spaces);

  dis_string_ptr += sprintf(dis_string_ptr,"%06X", (op_code & 0X00FFFFFF));

  }
return;
}

/*----------------------------------------------------------------------------*/

void dis_cc(int cc, int *spaces)
{
switch (cc)
  {
  case 0X0:dis_string_ptr += sprintf(dis_string_ptr,"EQ");break;
  case 0X1:dis_string_ptr += sprintf(dis_string_ptr,"NE");break;
  case 0X2:dis_string_ptr += sprintf(dis_string_ptr,"CS");break;
  case 0X3:dis_string_ptr += sprintf(dis_string_ptr,"CC");break;
  case 0X4:dis_string_ptr += sprintf(dis_string_ptr,"MI");break;
  case 0X5:dis_string_ptr += sprintf(dis_string_ptr,"PL");break;
  case 0X6:dis_string_ptr += sprintf(dis_string_ptr,"VS");break;
  case 0X7:dis_string_ptr += sprintf(dis_string_ptr,"VC");break;
  case 0X8:dis_string_ptr += sprintf(dis_string_ptr,"HI");break;
  case 0X9:dis_string_ptr += sprintf(dis_string_ptr,"LS");break;
  case 0XA:dis_string_ptr += sprintf(dis_string_ptr,"GE");break;
  case 0XB:dis_string_ptr += sprintf(dis_string_ptr,"LT");break;
  case 0XC:dis_string_ptr += sprintf(dis_string_ptr,"GT");break;
  case 0XD:dis_string_ptr += sprintf(dis_string_ptr,"LE");break;
  case 0XE:*spaces = *spaces + 2;break;
  case 0XF:dis_string_ptr += sprintf(dis_string_ptr,"NV");break;
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_tab(int spaces)
{
while (spaces > 0) {dis_string_ptr += sprintf(dis_string_ptr," "); spaces--;}
return;
}

/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*-------------------------THUMB disassemble ---------------------------------*/

void thumb_disassemble(unsigned int address, unsigned int op_code)
{
op_code = op_code & 0X0000FFFF;

switch (op_code & 0XE000)
  {
  case 0X0000: dis_data0(address, op_code);         break;
  case 0X2000: dis_data1(address, op_code);         break;
  case 0X4000: dis_data_transfer(address, op_code); break;
  case 0X6000: dis_transfer0(address, op_code);     break;
  case 0X8000: dis_transfer1(address, op_code);     break;
  case 0XA000: dis_sp_pc(address, op_code);         break;
  case 0XC000: dis_lsm_b(address, op_code);         break;
  case 0XE000: dis_thumb_branch(address, op_code);  break;
  }

return;
}

/*----------------------------------------------------------------------------*/

void case0_addr(unsigned int op_code, char *mnemonic)
{
int temp;

dis_string_ptr += sprintf(dis_string_ptr,"%s     R%d, R%d, #", mnemonic, op_code & 7, (op_code >> 3) & 7);
temp = (op_code>>6) & 0X1F;
if (((op_code & 0X1800) != 0) && (temp == 0)) temp = 32;     /* Special cases */
dis_string_ptr += sprintf(dis_string_ptr,"%02X", temp);
}

/*----------------------------------------------------------------------------*/

void dis_data0(unsigned int address, unsigned int op_code)
{
/*char mnemonic;              CB unused crying*/

switch(op_code & 0X1800)
  {
  case 0X0000: case0_addr(op_code, "LSL"); break;
  case 0X0800: case0_addr(op_code, "LSR"); break;
  case 0X1000: case0_addr(op_code, "ASR"); break;
  case 0X1800:
    if ((op_code & 0X0200) == 0) dis_string_ptr += sprintf(dis_string_ptr,"ADD"); else dis_string_ptr += sprintf(dis_string_ptr,"SUB");
    dis_string_ptr += sprintf(dis_string_ptr,"     R%d, R%d, ", op_code & 7, ((op_code >> 3) & 7));
    if ((op_code & 0X0400) == 0) dis_string_ptr += sprintf(dis_string_ptr,"R"); else dis_string_ptr += sprintf(dis_string_ptr,"#");
    dis_string_ptr += sprintf(dis_string_ptr,"%d", ((op_code>>6) & 7));
    break;
  }

return;
}


/*----------------------------------------------------------------------------*/

void dis_data1(unsigned int address, unsigned int op_code)
{
switch(op_code & 0X1800)
  {
  case 0X0000: dis_string_ptr += sprintf(dis_string_ptr,"MOV"); break;
  case 0X0800: dis_string_ptr += sprintf(dis_string_ptr,"CMP"); break;
  case 0X1000: dis_string_ptr += sprintf(dis_string_ptr,"ADD"); break;
  case 0X1800: dis_string_ptr += sprintf(dis_string_ptr,"SUB"); break;
  }

dis_string_ptr += sprintf(dis_string_ptr,"     R%d, #%02X", (op_code >> 8) & 7, op_code & 0XFF);

return;
}

/*----------------------------------------------------------------------------*/

void dis_data_transfer(unsigned int address, unsigned int op_code)
{
if ((op_code &0X1000) == 0)                                    /* load /store */
  {
  if ((op_code & 0X0800) == 0)
    {
    if ((op_code & 0X0400) == 0)                           /* data processing */
      {
      switch(op_code & 0X03C0)                      /* data processing opcode */
        {
        case 0X0000: dis_string_ptr += sprintf(dis_string_ptr,"AND"); break;
        case 0X0040: dis_string_ptr += sprintf(dis_string_ptr,"EOR"); break;
        case 0X0080: dis_string_ptr += sprintf(dis_string_ptr,"LSL"); break;
        case 0X00C0: dis_string_ptr += sprintf(dis_string_ptr,"LSR"); break;
        case 0X0100: dis_string_ptr += sprintf(dis_string_ptr,"ASR"); break;
        case 0X0140: dis_string_ptr += sprintf(dis_string_ptr,"ADC"); break;
        case 0X0180: dis_string_ptr += sprintf(dis_string_ptr,"SBC"); break;
        case 0X01C0: dis_string_ptr += sprintf(dis_string_ptr,"ROR"); break;
        case 0X0200: dis_string_ptr += sprintf(dis_string_ptr,"TST"); break;
        case 0X0240: dis_string_ptr += sprintf(dis_string_ptr,"NEG"); break;
        case 0X0280: dis_string_ptr += sprintf(dis_string_ptr,"CMP"); break;
        case 0X02C0: dis_string_ptr += sprintf(dis_string_ptr,"CMN"); break;
        case 0X0300: dis_string_ptr += sprintf(dis_string_ptr,"ORR"); break;
        case 0X0340: dis_string_ptr += sprintf(dis_string_ptr,"MUL"); break;
        case 0X0380: dis_string_ptr += sprintf(dis_string_ptr,"BIC"); break;
        case 0X03C0: dis_string_ptr += sprintf(dis_string_ptr,"MVN"); break;
        }
      dis_string_ptr += sprintf(dis_string_ptr,"     R%d, R%d", op_code & 7, (op_code>>3) & 7);
      }
    else                                            /*special data processing */
      {
      switch(op_code & 0X0300)
        {
        case 0X0000:
          dis_string_ptr += sprintf(dis_string_ptr,"ADD     R%d, R%d", ((op_code & 0X0080) >> 4) | (op_code & 7),
                                      (op_code >> 3) & 15);
          break;

        case 0X0100:
          dis_string_ptr += sprintf(dis_string_ptr,"CMP     R%d, R%d", ((op_code & 0X0080) >> 4) | (op_code & 7),
                                      (op_code >> 3) & 15);
          break;

        case 0X0200:
          dis_string_ptr += sprintf(dis_string_ptr,"MOV     R%d, R%d", ((op_code & 0X0080) >> 4) | (op_code & 7),
                                      (op_code >> 3) & 15);
            break;

        case 0X0300:                                                /* BX/BLX */
          if ((op_code & 0X0080) == 0) dis_string_ptr += sprintf(dis_string_ptr,"BX      R%d", (op_code>>3) & 15);
          else                         dis_string_ptr += sprintf(dis_string_ptr,"BLX     R%d", (op_code>>3) & 15);
          break ;
        }
      }                                     /* End of special data processing */
    }
  else                                              /* Load from literal pool */
    {
    dis_string_ptr += sprintf(dis_string_ptr,"LDR    R%d, [PC, #%03X]", (op_code >> 8) & 7,
           (op_code & 0X00FF) << 2);
    dis_string_ptr += sprintf(dis_string_ptr,"   (%08X)", (address & 0XFFFFFFFC) + 4 + ((op_code & 0x00FF) << 2));
    }
  }
else                           /* load/store word, halfword, byte, signed byte*/
  {
  switch(op_code & 0X0E00)
    {
    case 0X0000: dis_string_ptr += sprintf(dis_string_ptr,"STR   "); break;
    case 0X0200: dis_string_ptr += sprintf(dis_string_ptr,"STRH  "); break;
    case 0X0400: dis_string_ptr += sprintf(dis_string_ptr,"STRB  "); break;
    case 0X0600: dis_string_ptr += sprintf(dis_string_ptr,"LDRSB "); break;
    case 0X0800: dis_string_ptr += sprintf(dis_string_ptr,"LDR   "); break;
    case 0X0A00: dis_string_ptr += sprintf(dis_string_ptr,"LDRH  "); break;
    case 0X0C00: dis_string_ptr += sprintf(dis_string_ptr,"LDRB  "); break;
    case 0X0E00: dis_string_ptr += sprintf(dis_string_ptr,"LDRSH "); break;
    }

  dis_string_ptr += sprintf(dis_string_ptr,"  R%d, [R%d, R%d]", (op_code & 7), ((op_code >> 3) & 7),
                                             ((op_code >> 6) & 7));
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer0(unsigned int address, unsigned int op_code)
                                                  /* load/store w,b immediate */
{
if ((op_code & 0X0800) == 0) dis_string_ptr += sprintf(dis_string_ptr,"STR"); else dis_string_ptr += sprintf(dis_string_ptr,"LDR");
if ((op_code & 0X1000) == 0)
  dis_string_ptr += sprintf(dis_string_ptr,"    R%d, [R%d, #%02X]", op_code & 7, (op_code >> 3) & 7,
                                               (op_code >> 4) & 0X07C);
else
  dis_string_ptr += sprintf(dis_string_ptr,"B    R%d, [R%d, #%02X]", op_code & 7, (op_code >> 3) & 7,
                                                (op_code >> 6) & 0X1F);
return;
}

/*----------------------------------------------------------------------------*/

void dis_transfer1(unsigned int address, unsigned int op_code)
                                            /* load/store hw immediate & stack*/
{
if ((op_code & 0X0800) == 0) dis_string_ptr += sprintf(dis_string_ptr,"STR"); else dis_string_ptr += sprintf(dis_string_ptr,"LDR");

if ((op_code & 0X1000) == 0)
  dis_string_ptr += sprintf(dis_string_ptr,"H   R%d, [R%d, #%02X]", op_code & 7,
                                 (op_code >> 3) & 7,
                                 (op_code >> 5) & 0X3E);
else
  dis_string_ptr += sprintf(dis_string_ptr,"   R%d, [SP, #%03X]", (op_code >> 6) & 7, (op_code << 2) & 0X3FC);

return;
}

/*----------------------------------------------------------------------------*/

void dis_sp_pc(unsigned int address, unsigned int op_code)/* stack operations */
{
if ((op_code & 0X1000) == 0)                                  /* ADD SP or PC */
  {
  dis_string_ptr += sprintf(dis_string_ptr,"ADD    R%d,", (op_code >> 8) & 7);

  if ((op_code & 0X0800) == 0) dis_string_ptr += sprintf(dis_string_ptr,"PC"); else dis_string_ptr += sprintf(dis_string_ptr,"SP");
  dis_string_ptr += sprintf(dis_string_ptr,",  #%03X",  (op_code & 0X00FF) << 2);

  if ((op_code & 0X0800) == 0)
    dis_string_ptr += sprintf(dis_string_ptr,"  (%08X)", (address & 0XFFFFFFFC) + 4 + ((op_code & 0X00FF) << 2));
  }
else
  {
  switch(op_code & 0X0F00)
    {
    case 0X0000:                                                 /* Adjust SP */
      if ((op_code & 0X0080) == 0) dis_string_ptr += sprintf(dis_string_ptr,"ADD");
      else                         dis_string_ptr += sprintf(dis_string_ptr,"SUB");
      dis_string_ptr += sprintf(dis_string_ptr,"    SP, SP, #%03X", (op_code & 0X7F) << 2);
      break;

    case 0X0400:
    case 0X0500:
    case 0X0C00:
    case 0X0D00:
      if ((op_code & 0X0800) == 0) dis_string_ptr += sprintf(dis_string_ptr,"PUSH");
      else                         dis_string_ptr += sprintf(dis_string_ptr,"POP ");
      dis_string_ptr += sprintf(dis_string_ptr,"    {");
      print_reg_list(op_code & 0X00FF);

      if ((op_code & 0X0100) != 0)
        {
        if ((op_code & 0X00FF) != 0) dis_string_ptr += sprintf(dis_string_ptr,",");
        if ((op_code & 0X0800) == 0) dis_string_ptr += sprintf(dis_string_ptr,"LR");
        else                         dis_string_ptr += sprintf(dis_string_ptr,"PC");
        }
      dis_string_ptr += sprintf(dis_string_ptr,"}");
      break;

    case 0X0E00:
      dis_string_ptr += sprintf(dis_string_ptr,"Breakpoint");
      break;

    case 0X0100:
    case 0X0200:
    case 0X0300:
    case 0X0600:
    case 0X0700:
    case 0X0800:
    case 0X0900:
    case 0X0A00:
    case 0X0B00:
    case 0X0F00:
      dis_string_ptr += sprintf(dis_string_ptr,"Undefined");
      break;
    }
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_lsm_b(unsigned int address, unsigned int op_code)
{
if ((op_code & 0X1000) == 0)
  {
  if ((op_code & 0X0800) == 0) dis_string_ptr += sprintf(dis_string_ptr,"STM"); else dis_string_ptr += sprintf(dis_string_ptr,"LDM");
  dis_string_ptr += sprintf(dis_string_ptr,"IA   R%d", (op_code >> 8) & 7);

                                           /*option for writeback suppression */
  if (((op_code & 0X0800) == 0)
    || ((1 << ((op_code >> 8) & 7) & op_code) == 0)) dis_string_ptr += sprintf(dis_string_ptr,"!");

  dis_string_ptr += sprintf(dis_string_ptr,", {"); print_reg_list(op_code & 0X00FF); dis_string_ptr += sprintf(dis_string_ptr,"}");
  }
else
  {
  if ((op_code & 0X0F00) != 0X0F00)
    {
    int offset;

    switch (op_code & 0X0F00)
      {
      case 0X0000: dis_string_ptr += sprintf(dis_string_ptr,"BEQ     "); break;
      case 0X0100: dis_string_ptr += sprintf(dis_string_ptr,"BNE     "); break;
      case 0X0200: dis_string_ptr += sprintf(dis_string_ptr,"BCS     "); break;
      case 0X0300: dis_string_ptr += sprintf(dis_string_ptr,"BCC     "); break;
      case 0X0400: dis_string_ptr += sprintf(dis_string_ptr,"BMI     "); break;
      case 0X0500: dis_string_ptr += sprintf(dis_string_ptr,"BPL     "); break;
      case 0X0600: dis_string_ptr += sprintf(dis_string_ptr,"BVS     "); break;
      case 0X0700: dis_string_ptr += sprintf(dis_string_ptr,"BVC     "); break;
      case 0X0800: dis_string_ptr += sprintf(dis_string_ptr,"BHI     "); break;
      case 0X0900: dis_string_ptr += sprintf(dis_string_ptr,"BLS     "); break;
      case 0X0A00: dis_string_ptr += sprintf(dis_string_ptr,"BGE     "); break;
      case 0X0B00: dis_string_ptr += sprintf(dis_string_ptr,"BLT     "); break;
      case 0X0C00: dis_string_ptr += sprintf(dis_string_ptr,"BGT     "); break;
      case 0X0D00: dis_string_ptr += sprintf(dis_string_ptr,"BLE     "); break;
      case 0X0E00: dis_string_ptr += sprintf(dis_string_ptr,"B       "); break;
      case 0X0F00: dis_string_ptr += sprintf(dis_string_ptr,"wrong !");  break;
      }

    offset = (op_code & 0X00FF) << 1;
    if ((op_code & 0X0080) != 0) offset = offset | 0XFFFFFE00;
    dis_string_ptr += sprintf(dis_string_ptr,"%08X", address + 4 + offset);
    }
  else dis_string_ptr += sprintf(dis_string_ptr,"SWI     #%02X", op_code & 0XFF);
  }

return;
}

/*----------------------------------------------------------------------------*/

void dis_thumb_branch1(unsigned int op_code, int exchange)
{
int offset;

offset = BL_prefix << 12;
if ((BL_prefix & 0X0400) != 0) offset = offset | 0XFF800000;
offset = offset | ((op_code & 0X07FF) << 1);
if (exchange != TRUE) dis_string_ptr += sprintf(dis_string_ptr,"BL      %08X", BL_address + offset);
else dis_string_ptr += sprintf(dis_string_ptr,"BLX     %08X", (BL_address + offset) & 0XFFFFFFFC);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void dis_thumb_branch(unsigned int address, unsigned int op_code)
{
switch (op_code & 0X1800)
  {
  case 0X0000: dis_string_ptr += sprintf(dis_string_ptr,"B   ");
    {
    int offset;
    offset = (op_code & 0X07FF) << 1;
    if ((op_code & 0X0400) != 0) offset = offset | 0XFFFFF000;
    dis_string_ptr += sprintf(dis_string_ptr,"%08X", address + 4 + offset);
    }
    break;

  case 0X0800:
    if ((op_code & 0X0001) == 0X0000) dis_thumb_branch1(op_code, TRUE);
    else dis_string_ptr += sprintf(dis_string_ptr,"Undefined");
    break;

  case 0X1000: dis_string_ptr += sprintf(dis_string_ptr,"BL prefix");
    BL_prefix  = op_code & 0X07FF;
    BL_address = address + 4;
    break;

  case 0X1800:
    dis_thumb_branch1(op_code, FALSE);
    break;
  }

return;
}

