/*		Name:		asm.c
		Version:	0.0.12
		Date:		23/6/2001
		(Assembly and) Dissassembly functions 
*/

/* This code is property of Charles Brej  and is used in this project 
									with his permission */
#include "asm.h"
#include "dotparse.h"
#include <glib.h>
#include <ctype.h>
#include <stdint.h>

DefinitionStack *asm_parser (GList * list)
{
 
	if (!SCAN_IS_STRING (SCAN_LIST_FIRST (list)))
	{
		g_print ("line:%d position:%d  Expected string to define the disassembly\n",
			SCAN_LIST_FIRST (list)->line, SCAN_LIST_FIRST (list)->position);
		return NULL;
	}
	return asm_define (list->next, NULL, SCAN_STRING (SCAN_NODE (list->data)));
}

DefinitionStack *asm_define (GList * parselist, DefinitionStack * deflist, char *name)
{
	DefinitionStack *definitionelement = g_new (DefinitionStack, 1);
	DefinitionStack *definitionstack = definitionelement;
	DefinitionStack *deflisttemp;
	GList *sublist;

	definitionelement->next = deflist;
	definitionelement->rules = NULL;
	definitionelement->string = name;
	while (parselist)
	{
		if (SCAN_IS_LIST ((PtrSCANNode) parselist->data))
		{
			sublist = ((PtrSCANNode) parselist->data)->body.list;
			if (!sublist)
			{
				g_print ("line:%d position:%d  Empty list when defining '%s'\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position, name);
				return definitionelement;
			}

			if (!sublist->next)
			{
				g_print ("line:%d position:%d  List of only 1 element when defining '%s'\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position, name);
				return definitionelement;
			}

			if (SCAN_IS_SYMBOL ((PtrSCANNode) sublist->data) &&
				SCAN_IS_STRING ((PtrSCANNode) sublist->next->data) &&
				!g_strcasecmp (((PtrSCANNode) sublist->data)->body.string, DEFINITIONSTRING))
				definitionstack = asm_define (sublist->next->next, definitionstack,
					((PtrSCANNode) sublist->next->data)->body.string);
			else
				definitionelement->rules = g_list_append (definitionelement->rules,
					(gpointer) (asm_rule (sublist, definitionstack)));



		}

		if (SCAN_IS_SYMBOL ((PtrSCANNode) parselist->data))
		{
			deflisttemp = definitionstack;
			while (g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, deflisttemp->string))
			{
				deflisttemp = deflisttemp->next;
				if (!deflisttemp)
				{
					g_print ("line:%d position:%d  '%s' has not been defined\n",
						((PtrSCANNode) parselist->data)->line,
						((PtrSCANNode) parselist->data)->position, ((PtrSCANNode) parselist->data)->body.string);
					return definitionelement;
				}
			}
			definitionelement->rules = g_list_concat (definitionelement->rules, g_list_copy (deflisttemp->rules));
		}
		parselist = parselist->next;
	}

	/*shuld free all elements above the definition stack */
	deflisttemp = definitionstack;
	while (definitionstack != definitionelement)
	{
		deflisttemp = definitionstack;
		definitionstack = definitionstack->next;
		g_free (deflisttemp);
	}

	return definitionelement;
}

Rule *asm_rule (GList * parselist, DefinitionStack * deflist)
{
	Rule *rule = g_new (Rule, 1);
	RuleElement *belement, *telement;
	GList *text;
	GList *binary;
	DefinitionStack *deflisttemp;
	GList *binarytemp;
	char *legaltest;

	rule->asci = NULL;
	rule->bit = NULL;
	rule->type = TRANRULE;
	if (SCAN_IS_SYMBOL ((PtrSCANNode) parselist->data))
	{
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, INTIGERSTRING))
			rule->type = INTRULE;
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, RELATIVESTRING))
			rule->type = RELATIVERULE;
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, UINTIGERSTRING))
			rule->type = UINTRULE;
		if (!g_strcasecmp (((PtrSCANNode) parselist->data)->body.string, URELATIVESTRING))
			rule->type = URELATIVERULE;
		if (TRANRULE == rule->type)
		{
			g_print ("line:%d position:%d  Only 'int' and 'relative' are valid rules \n",
				((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
			return rule;
		}
		parselist = parselist->next;

		if (!SCAN_IS_NUMBER ((PtrSCANNode) parselist->data))
		{
			g_print ("line:%d position:%d  Int and float rules need a size as the first operand \n",
				((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
			return rule;
		}

		rule->bit = g_list_append (rule->bit, GINT_TO_POINTER (((PtrSCANNode) parselist->data)->body.number));
		parselist = parselist->next;
		while (parselist)
		{
			if (!parselist->next)
			{
				g_print
					("line:%d position:%d  Int and float rules need mathematical operations in sepperated pairs eg: + 4 * 8\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
				return rule;
			}

			if (!SCAN_IS_SYMBOL ((PtrSCANNode) parselist->data) ||
				!SCAN_IS_NUMBER ((PtrSCANNode) parselist->next->data))
			{
				g_print
					("line:%d position:%d  Int and float rules need mathematical operations in sepperated pairs eg: + 4 * 8\n",
					((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
				return rule;
			}

			rule->asci = g_list_append (rule->asci, ((PtrSCANNode) parselist->data)->body.string);
			rule->bit = g_list_append (rule->bit, GINT_TO_POINTER (((PtrSCANNode) parselist->next->data)->body.number));
			parselist = parselist->next->next;
		}
		return rule;
	}


	rule->type = TRANRULE;
	if (parselist->next->next ||
		!SCAN_IS_LIST ((PtrSCANNode) parselist->data) || !SCAN_IS_LIST ((PtrSCANNode) parselist->next->data))
	{
		g_print ("line:%d position:%d  A rule must have two LISTS! \n",
			((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
		return rule;
	}
	text = SCAN_NODE (parselist->data)->body.list;
	binary = SCAN_NODE (parselist->next->data)->body.list;
	while (binary)
	{
		belement = g_new (RuleElement, 1);
		if (!SCAN_IS_SYMBOL (SCAN_NODE (binary->data)))
		{
			g_print
				("line:%d position:%d  A rule binary field must be made from characters IOXZ or a number starting with '#'\n",
				((PtrSCANNode) binary->data)->line, ((PtrSCANNode) binary->data)->position);
			return rule;
		}
		belement->type = TERM;
		belement->string = SCAN_NODE (binary->data)->body.string;
		rule->bit = g_list_append (rule->bit, (gpointer) belement);
		binary = binary->next;
	}

	while (text)
	{
		telement = g_new (RuleElement, 1);
		rule->asci = g_list_append (rule->asci, (gpointer) telement);

		switch (((PtrSCANNode) text->data)->type)
		{
		case SCANString:
			telement->string = SCAN_NODE (text->data)->body.string;
			telement->type = TERM;
			break;

		case SCANList:
			if (!SCAN_NODE (text->data)->body.list)
			{
				g_print ("line:%d position:%d  Rule must have text body \n",
					((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position);
				return rule;
			}
			if (SCAN_IS_SYMBOL (SCAN_NODE (SCAN_NODE (text->data)->body.list->data)))
			{
				g_print ("line:%d position:%d  '%s' shuld be a in quotes\n",
					((PtrSCANNode) text->data)->line,
					((PtrSCANNode) text->data)->position,
					(SCAN_NODE (SCAN_NODE (text->data)->body.list->data))->body.string);
				return rule;
			}

			if (!SCAN_IS_STRING (SCAN_NODE (SCAN_NODE (text->data)->body.list->data)))
			{
				g_print ("line:%d position:%d  no name for inline definition",
					((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position);
				return rule;
			}


			telement->string = SCAN_NODE (SCAN_NODE (text->data)->body.list->data)->body.string;
			deflisttemp = asm_define (SCAN_NODE (text->data)->body.list->next, deflist, telement->string);

			binarytemp = rule->bit;
			while (g_strcasecmp (telement->string, ((RuleElement *) binarytemp->data)->string))
			{
				if (!binarytemp->next)
				{
					g_print ("line:%d position:%d  '%s' needs to be in the binary section too\n",
						((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
					return rule;
				}
				binarytemp = binarytemp->next;

			}
			((RuleElement *) binarytemp->data)->type = NONTERM;
			telement->type = NONTERM;
			((RuleElement *) binarytemp->data)->rules = deflisttemp->rules;
			telement->rules = deflisttemp->rules;
			((RuleElement *) binarytemp->data)->twin = g_list_index (rule->asci, telement);
			telement->twin = g_list_position (rule->bit, binarytemp);
			g_free (deflisttemp);
			break;

		case SCANSymbol:
			telement->string = SCAN_NODE (text->data)->body.string;
			deflisttemp = deflist;
			while (g_strcasecmp (telement->string, deflisttemp->string))
			{
				deflisttemp = deflisttemp->next;
				if (!deflisttemp)
				{
					g_print ("line:%d position:%d  '%s' has not been defined\n",
						((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
					return rule;
				}
			}

			binarytemp = rule->bit;
			while (g_strcasecmp (telement->string, ((RuleElement *) binarytemp->data)->string))
			{
				binarytemp = binarytemp->next;
				if (!binarytemp)
				{
					g_print ("line:%d position:%d  '%s' needs to be in the binary section too\n",
						((PtrSCANNode) text->data)->line, ((PtrSCANNode) text->data)->position, telement->string);
					return rule;
				}
			}
			((RuleElement *) binarytemp->data)->type = NONTERM;
			telement->type = NONTERM;
			((RuleElement *) binarytemp->data)->rules = deflisttemp->rules;
			telement->rules = deflisttemp->rules;
			((RuleElement *) binarytemp->data)->twin = g_list_index (rule->asci, telement);
			telement->twin = g_list_position (rule->bit, binarytemp);
			break;
		default:
			g_print ("line:%d position:%d  unexpected thing!\n",
				((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position);
			return rule;
		}
		text = text->next;
	}
	binary = rule->bit;
	while (binary)
	{
		if (TERM == ((RuleElement *) binary->data)->type)
		{
			legaltest = ((RuleElement *) binary->data)->string;
			if ('#' == *legaltest)
			{
				legaltest++;
				((RuleElement *) binary->data)->string = legaltest;
			}
			while (*legaltest)
			{
				if ('1' == *legaltest)
					*legaltest = 'I';
				if ('0' == *legaltest)
					*legaltest = 'O';
				legaltest++;
			}
			legaltest = ((RuleElement *) binary->data)->string;
			while (*legaltest)
			{
				if ('I' != *legaltest && 'O' != *legaltest && 'Z' != *legaltest && 'X' != *legaltest)
				{
					g_print
						("line:%d position:%d  This rule's binary field contains '%s' which is not a valid binary value nor is it in the text field\n",
						((PtrSCANNode) parselist->data)->line, ((PtrSCANNode) parselist->data)->position,
						((RuleElement *) binary->data)->string);
					break;
				}
				legaltest++;
			}
		}
		binary = binary->next; 
	}
	return rule;
}

Disasm asm_disassemble (GList * bitfield, GList * rules, int address)
{
	Disasm ret;
	Disasm rep;
	Disasm error;
	GList *binary;
	char *c;
	RuleElement *newelement;
	unsigned int flag;
    unsigned int size;
    
	error.distext = &dis_error;
	error.bitfield = &dis_error;

	while (rules)
	{
		ret.bitfield = bitfield;
		if (TRANRULE != ((Rule *) rules->data)->type)
		{   if ((int) ((Rule *) rules->data)->bit->data > g_list_length (bitfield)) return error;
			binary = g_list_nth (ret.bitfield, (int) ((Rule *) rules->data)->bit->data - 1);
			flag = 0;
			if (GPOINTER_TO_INT (binary->data))
                if (    UINTRULE      == ((Rule *) rules->data)->type
                     || URELATIVERULE == ((Rule *) rules->data)->type)
				        flag=1;
                else    flag=-1;
			while (binary != ret.bitfield)
			{
				binary = binary->prev;
				flag = flag << 1;
				if (GPOINTER_TO_INT (binary->data))
					flag++;
			}
            {
                GList *number = g_list_last (((Rule *) rules->data)->bit);
                GList *opp    = g_list_last (((Rule *) rules->data)->asci);
                int imm;
                    while(opp){
                        imm=GPOINTER_TO_INT (number->data);
                        c= (char*)opp->data;
                        if ('~' == *c) {
                            imm = flag;
                            flag = (int) number->data;
                            c++;
                            }
                        switch (*c){
                            case '+':
                                flag -=imm;
                                if (    UINTRULE      == ((Rule *) rules->data)->type
                                     || URELATIVERULE == ((Rule *) rules->data)->type)
                                                    if ((int)flag<0) return error;
                                break;
                    
                            case '-':
                                flag +=imm;
                                break;
            
                            case '*':
                                flag /=imm;
                                break;
                    
                            case '/':
                                flag *=imm;
                                break;
            
                            case '>':
                                flag = flag << imm;
                                break;
                                
                            case '<':
                                flag = flag >> imm;
                                break;
                                
                            case '@':
                                c++;
                                size = g_strtod(c,NULL);
                                flag = (((flag >> imm) | (flag << (size - imm))) & ((((1 << (size-1) )-1)<<1)+1 ));
                                break;
                                
                            default:
                                break;
                            }
                        opp=opp->prev;
                        number=number->prev;
                        }
            }
			if (    (RELATIVERULE == ((Rule *) rules->data)->type)
                || (URELATIVERULE == ((Rule *) rules->data)->type))
				flag += address;
			ret.bitfield = g_list_nth (ret.bitfield, (int) ((Rule *) rules->data)->bit->data);
			newelement = g_new (RuleElement, 1);
			newelement->string = g_strdup_printf ("%X", flag);
			newelement->type = TERM2FREE;
			ret.distext = g_list_append (NULL, newelement);
			return ret;
		}

		binary = g_list_last (((Rule *) rules->data)->bit);
		ret.distext = g_list_copy (((Rule *) rules->data)->asci);

		while (1)
		{
			flag = 0;
			if (!binary) return ret;
			if (NONTERM == ((RuleElement *) binary->data)->type)
			{
				rep = asm_disassemble (ret.bitfield, ((RuleElement *) binary->data)->rules, address);
				if (&dis_error != rep.bitfield)
				{
					ret.bitfield = rep.bitfield;
					newelement = g_new (RuleElement, 1);
					newelement->rules = rep.distext;
					newelement->type = NONTERM2FREE;
					(g_list_nth (ret.distext, ((RuleElement *) binary->data)->twin))->data = (gpointer) newelement;
				} else
					flag = 1;
			} else
			{
				c = (char *) ((RuleElement *) binary->data)->string;
				while (*c)
					c++;
                if (ret.bitfield)
				 while (((RuleElement *) binary->data)->string != c)
				 {
					c--;
					if (ret.bitfield){
                     if ('I' == *c && !GPOINTER_TO_INT (ret.bitfield->data))
						 flag = 1;
					 if ('O' == *c && GPOINTER_TO_INT (ret.bitfield->data))
						 flag = 1;
					 ret.bitfield = ret.bitfield->next;
                     }
                    else flag = 1;
				 }
                else flag = 1;
                 
		    }
		if (flag)
			break;
		binary = binary->prev;
        }
	asm_freestringlist (ret.distext);
	rules = rules->next;
	}
	return error;
}

GList* asm_assemble (char* text, GList* rules, int address, int last)
{
 Asm *ret = g_new(Asm,1);
 Asm *todo;
 GList *todolist=NULL;
 GList *retlist=NULL;
 GList *rep;
 RuleElement *newelement;
 RuleElement *newelement2;
 int flag;
 unsigned int size;
 unsigned int value;
 char* start;
 char* c;
 GList* test;
    
 while (rules){
        flag=1;
	    ret->text = text;
	    if (TRANRULE != ((Rule *) rules->data)->type){
            while (' '  == *ret->text) ret->text++;
            if (!isxdigit(*ret->text))    break;
            flag=0;
            value=0;
            while (isxdigit(*ret->text)){
                   if (*ret->text< 0x3A)    value = (value<<4) + *ret->text - 0x30;
                   else                     value = (value<<4) + (*ret->text & 7) + 9;
                   ret->text++;
                   }
			if (    (RELATIVERULE == ((Rule *) rules->data)->type)
                || (URELATIVERULE == ((Rule *) rules->data)->type))
				    value -= address;
            {
                GList *number = ((Rule *) rules->data)->bit->next;
                GList *opp = ((Rule *) rules->data)->asci;
                int imm;
                    while(opp){
                        imm=GPOINTER_TO_INT (number->data);
                        c= (char*)opp->data;
                        
                        if ('~' == *c) {
                            imm = value;
                            value = (int) number->data;
                            c++;
                            }
                            
                        switch (*c){
                            case '+':
                                value +=imm;
                                break;
                    
                            case '-':
                                value -=imm;
                                break;
            
                            case '*':
                                value *=imm;
                                break;
                    
                            case '/':
                                value /=imm;
                                break;
            
                            case '>':
                                c++;
                                value = value >> imm;
                                break;
                                
                            case '<':
                                value = value << imm;
                                break;
                                
                            case '@':
                                c++;
                                size = g_strtod(c,NULL);
                                value = ((value << imm) | (value >> (size - imm)) & ((((1 << (size-1) )-1)<<1)+1 ));
                                break;
                                
                            default:
                                break;
                            }
                        opp=opp->next;
                        number=number->next;
                        }
            }
            if ((value < (1<<(unsigned int) ((Rule *) rules->data)->bit->data))
                || (RELATIVERULE  == ((Rule *) rules->data)->type)
                || (URELATIVERULE == ((Rule *) rules->data)->type))
            if (!flag) {
                start = g_new(char,(int) ((Rule *) rules->data)->bit->data + 1);
                c= start + (int) ((Rule *) rules->data)->bit->data;
                *c='\0';
                while(start != c--){
                    if (value & 1)  *c = 'I';
                    else            *c = 'O';
                    value = value >> 1;
                    }
	    		newelement = g_new (RuleElement, 1);
    			newelement->string = start;
			    newelement->type = TERM2FREE;
		    	ret->binary = g_list_append (NULL, newelement);
                if (last) while (' '  == *ret->text) ret->text++;
                if (!(last && *(ret->text))) {
                    retlist = g_list_append(retlist,ret);
                    ret = g_new(Asm,1);
                    }
                }
		    }
        else {
	      ret->text = text;
          ret->ruletext=((Rule *) rules->data)->asci;
		  ret->binary = g_list_copy (((Rule *) rules->data)->bit);
        
		  while (1){
		    flag = 0;
            while (' '  == *(ret->text)) ret->text++;
		    if (!ret->ruletext) if (last && *(ret->text)) {asm_freestringlist (ret->binary);break;}
                                else {retlist = g_list_append(retlist,ret); ret = g_new(Asm,1);break;}
            
            
		    if (NONTERM == ((RuleElement *) ret->ruletext->data)->type){
                while (' '  == *(ret->text)) ret->text++;
				rep = asm_assemble(ret->text, ((RuleElement *) ret->ruletext->data)->rules, address, last && !ret->ruletext->next);

 				if (rep){
                    ret->text = ((Asm*)rep->data)->text;
 					newelement = g_new (RuleElement, 1);
 					newelement->rules = ((Asm*)rep->data)->binary;
 					newelement->type = NONTERM2FREE;
                    rep = rep->next;
                    while (rep) {
     					newelement2 = g_new (RuleElement, 1);
 	    				newelement2->rules = ((Asm*)rep->data)->binary;
 		    			newelement2->type = NONTERM2FREE;
                        todo=rep->data;
                        todo->ruletext=ret->ruletext;
 					    (g_list_nth (todo->binary=asm_copystringlist(ret->binary), ((RuleElement *) todo->ruletext->data)->twin))->data = (gpointer) newelement2;
                        todolist=g_list_prepend(todolist,todo);
                        rep=rep->next;
                        }
 					(g_list_nth (ret->binary, ((RuleElement *) ret->ruletext->data)->twin))->data = (gpointer) newelement;
				    }
                else  flag = 1; 
			    }
            else {
				c = (char *) ((RuleElement *) ret->ruletext->data)->string;
                    while (' '  == *c) c++;
                    while (' '  == *ret->text) ret->text++;
                if ('\t' != *c)
                  while (!flag){
                    while (' '  == *c) c++;
                    while (' '  == *ret->text) ret->text++;
                    if (!*c) break;
                    if (*(ret->text) != *c){
                        if ( (*c        > 0x40) &&
                             (*ret->text > 0x40) &&
                             ((*c        & 0x1f) < 0x1b) &&
                             ((*ret->text & 0x1f) < 0x1b) &&
                             (*c & 0x1f) == (*ret->text & 0x1f)  ){
                                c++;
                                ret->text++;
                                }
                        else flag=1;
                        }
                    else { c++; ret->text++; }
                    }
                else {
                    if ('f'==*(c+1)) 
                        if (' ' != *ret->text && ' ' != *(ret->text-1))
                            flag=1;
                     }
                }
			if (flag)
                if (todolist) {
                    g_free(ret);
                    ret = todolist->data;
                    todolist=g_list_remove(todolist, todolist->data);
                    }
                else break;
			ret->ruletext = ret->ruletext ->next;
                
            }
            
          }
        rules=rules->next;
	    }
	return retlist;
}



















Asm asm_assembleold (char* text, GList* rules, int address, int last, int count)
{
 Asm error;
 Asm ret;
 Asm rep;
 GList *ruletext;
 RuleElement *newelement;
 int flag;
 int size;
 char* start;
 char* c;
 GList* tried = NULL;
 GList* current = NULL;    
    
 error.text = (char*) &dis_error;
 error.binary = &dis_error;

while (rules){
	    ret.text = text;
        flag=1;
	    if (TRANRULE != ((Rule *) rules->data)->type){
            while (' '  == *ret.text) ret.text++;
            if (!isxdigit(*ret.text))    break;
            if (count)    break;
            flag=0;
            while (isxdigit(*ret.text)){
                   if (*ret.text< 0x3A)     flag = (flag<<4) + *ret.text - 0x30;
                   else                     flag = (flag<<4) + (*ret.text & 7) + 9;
                   ret.text++;
                   }
			if (    (RELATIVERULE == ((Rule *) rules->data)->type)
                || (URELATIVERULE == ((Rule *) rules->data)->type))
				    flag -= address;
            {
                GList *number = ((Rule *) rules->data)->bit->next;
                GList *opp = ((Rule *) rules->data)->asci;
                int imm;
                    while(opp){
                        imm=GPOINTER_TO_INT (number->data);
                        c= (char*)opp->data;
                        if ('~' == *c) {
                            imm = flag;
                            flag = (int) number->data;
                            c++;
                            }
                        switch (*c){
                            case '+':
                                flag +=imm;
                                break;
                    
                            case '-':
                                flag -=imm;
                                break;
            
                            case '*':
                                flag *=imm;
                                break;
                    
                            case '/':
                                flag /=imm;
                                break;
            
                            case '>':
                                flag = flag >> imm;
                                break;
                                
                            case '<':
                                flag = flag << imm;
                                break;
                                
                            case '@':
                                flag = ((flag << imm) | (flag >> (int)((Rule *) rules->data)->bit->data - imm)) & (1 << (int)((Rule *) rules->data)->bit->data);
                                break;
                                
                            default:
                                break;
                            }
                        opp=opp->next;
                        number=number->next;
                        }
            }
            
            start = g_new(char,(int) ((Rule *) rules->data)->bit->data);
            c= start + (int) ((Rule *) rules->data)->bit->data;
            *c='\0';
            while(start != c--){
                if (flag & 1)   *c = 'I';
                else            *c = 'O';
                flag = flag >> 1;
                }
			newelement = g_new (RuleElement, 1);
			newelement->string = start;
			newelement->type = TERM2FREE;
			ret.binary = g_list_append (NULL, newelement);
            
            while (' '  == *ret.text) ret.text++;
            if (last && *ret.text)  break;
            else                    return ret;
		    }
        else {
	      ret.text = text;
          current = tried;
          ruletext=((Rule *) rules->data)->asci;
		  ret.binary = g_list_copy (((Rule *) rules->data)->bit);
        
		  while (1){
		    flag = 0;
            while (' '  == *ret.text) ret.text++;
		    if (!ruletext) 
                if (last && *ret.text)  break;
                else        
                    if (!count--)   return ret;
                    else    {
 	                    ret.text = text;
                        current = tried;
                        ruletext=((Rule *) rules->data)->asci;
		                ret.binary = g_list_copy (((Rule *) rules->data)->bit);
                        break;
                        }
            
            
		    if (NONTERM == ((RuleElement *) ruletext->data)->type){
                if (!current){
                    tried = g_list_append (tried,(int) 0);
                    current = g_list_last(tried);
				    rep = asm_assembleold (ret.text, ((RuleElement *) ruletext->data)->rules, address, last && !ruletext->next,(int) current->data);
                    }
                else {
                    if (!current->next) ((int)current->data)++;
                    else if (-1 == (int)current->next->data) {((int)current->data)++;(int)current->next->data=0;}
                    rep = asm_assembleold (ret.text, ((RuleElement *) ruletext->data)->rules, address, last && !ruletext->next,(int) current->data);
                    }
                
                    
 				if (&dis_error != rep.binary){
                    ret.text = rep.text;
					newelement = g_new (RuleElement, 1);
					newelement->rules = rep.binary;
					newelement->type = NONTERM2FREE;
					(g_list_nth (ret.binary, ((RuleElement *) ruletext->data)->twin))->data = (gpointer) newelement;
				    }
                else {  
                    flag = 1; 
                    ((int)current->data) = -1;
                    }
               current=current ->next;
			    }
            else {
				c = (char *) ((RuleElement *) ruletext->data)->string;
                if ('\t' != *c) while (!flag){
                    while (' '  == *c) c++;
                    while (' '  == *ret.text) ret.text++;
                    if (!*c) break;
                    if (*(ret.text) != *c){
                        if ( (*c        > 0x40) &&
                             (*ret.text > 0x40) &&
                             ((*c        & 0x1f) < 0x1b) &&
                             ((*ret.text & 0x1f) < 0x1b) &&
                             (*c & 0x1f) == (*ret.text & 0x1f)  ){
                                c++;
                                ret.text++;
                                }
                        else flag=1;
                        }
                    else { c++; ret.text++; }
                    }
                    
                }
			if (flag) break;
			ruletext = ruletext ->next;
            }
		  asm_freestringlist (ret.binary);
          }
	    if (NULL == tried  || -1 == (int) tried->data){
                    rules = rules->next;
                    tried = NULL; 
                    current = NULL;
                    }

	    }
	return error;
}


GList* asm_copystringlist (GList* list)
{
 GList *scan;
 RuleElement* newrule;
 list = g_list_copy(list);
 scan = list;
 while (scan && &dis_error != scan)
	{
		if (NONTERM2FREE == ((RuleElement *) scan->data)->type)
		{
            newrule = g_new(RuleElement, 1);
            newrule->type = NONTERM2FREE;
            newrule->string = ((RuleElement *) scan->data)->string;
            newrule->rules = asm_copystringlist (((RuleElement *) scan->data)->rules);
            scan->data = newrule;
		}
		if (TERM2FREE == ((RuleElement *) scan->data)->type)
		{
            newrule = g_new(RuleElement, 1);
            newrule->type = TERM2FREE;
            newrule->string =  g_strdup(((RuleElement *) scan->data)->string);
            newrule->rules = NULL;
            scan->data = newrule;
        }
		scan = scan->next;
	}
    return list;
}

void asm_freestringlist (GList* list)
{
	GList *scan = list;

	while (scan && &dis_error != scan)
	{
		if (NONTERM2FREE == ((RuleElement *) scan->data)->type)
		{
			asm_freestringlist (((RuleElement *) scan->data)->rules);
			g_free (scan->data);
		}
		if (TERM2FREE == ((RuleElement *) scan->data)->type)
		{
			g_free (((RuleElement *) scan->data)->string);
			g_free (scan->data);
		}
		scan = scan->next;
	}

	g_list_free (list);
}


GList *uchararr2glist (int size, uchar * vals)
{
	GList *ret = NULL;
	uchar c;
	int count;

	while (size--)
	{
		c = *vals;
		vals++;
		for (count = 8; count; count--)
		{
			ret = g_list_append (ret, GINT_TO_POINTER (c & 1));
			c >>= 1;
		}
	}
	return ret;
}

void asm_print (GList* list)
{
	if (&dis_error == list)
		g_print ("Undefined");
	else
		while (list)
		{
			if (TERM == ((RuleElement *) list->data)->type || 
                TERM2FREE == ((RuleElement *) list->data)->type)
				g_print ("%s", ((RuleElement *) list->data)->string);
			else
				asm_print (((RuleElement *) list->data)->rules);
			list = list->next;
		}

}

char* asm_sprint (GList* list, char* replystring)
{
 char* oldreplystring;
 int value;
		while (list)
		{
		if (TERM == ((RuleElement *) list->data)->type || 
                    TERM2FREE == ((RuleElement *) list->data)->type) {
                oldreplystring = replystring;
                replystring = asm_sprint_tabstring(((RuleElement *)list->data)->string, replystring);
                if (*oldreplystring)g_free(oldreplystring);
                }
		else    replystring = asm_sprint (((RuleElement *) list->data)->rules, replystring);
		list = list->next;
		}
/*                  Replaced by usng the tabstring function
 *  oldreplystring = replystring;
 *  while (oldreplystring)
 *         if ('\\' == *oldreplystring && 't' == *(oldreplystring+1)) {
 *                 char* temp = replystring;
 *                 value=0;
 *                 *oldreplystring='\0';
 *                 oldreplystring += 2;
 *                 while ( '0' >= *oldreplystring &&
 *                         '9' <= *oldreplystring ){
 *                             value = value * 10 + (*oldreplystring - '0');
 *                             oldreplystring++;
 *                             }
 *                 if (*oldreplystring) oldreplystring++;
 *                 g_strlen(replystring);
 *                 }
 */
                
                

 return replystring;
}

char* asm_sprint_tabstring (char* tabstring, char* replystring)
{
 int value;
 int value2=0;
 char* temp;
 if ('\t' != *tabstring)  return g_strconcat(replystring, tabstring, NULL);
 if ('f' == *(tabstring+1)) if (*(tabstring+2)) value2 = g_strtod (tabstring+2, &temp);
 else                       if (*(tabstring+2)) value2 = g_strtod (tabstring+1, &temp);
 value = strlen(replystring);
 if (value2) value = value2 - (value % value2);
 else value = 0;
 if (value<=0) return g_strconcat(replystring, tabstring, NULL);
 temp = g_new(char, value + 1);
 temp[value] = '\0';
 while (value) temp[--value] = ' ';
 replystring = g_strconcat(replystring, temp, NULL);
 g_free (temp);
 return replystring;
}


int asm_sbprint (GList* list, uchar** data)
{
    char* text = asm_sprint (list, "");
    int size = strlen(text);
    char* c = text + size;
    if (size%8) size = (size>>3) + 1;
    else        size =  size>>3;
    *data = g_new0(char, size);
    size = 0;
    while (c-->text){
        if ('X'==*c  || 'I'==*c) ((*data)[size>>3]) |= 1 << (size&7);
        size++;
        }
    return (size>>3);
}




