/*
 * disasm - a MIPS disassembler.
 *
 * Copyright (C) 1995-1996 Kai Harrekilde-Petersen <khp@dolphinics.no>
 *
 * Version: 0.38
 * Date: 22. August 1996
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "mips.h"

#define VERSION "0.38"

char *mips_regs[32] = {
  "zero", "at", "v0", "v1",
  "a0",   "a1", "a2", "a3",
  "t0",   "t1", "t2", "t3",
  "t4",   "t5", "t6", "t7",
  "s0",   "s1", "s2", "s3",
  "s4",   "s5", "s6", "s7",
  "t8",   "t9", "k0", "k1",
  "gp",   "sp", "s8", "ra"
};

char *algor_regs[32] = {
"$0" ,"$1" ,"$2"  ,"$3"  ,"$4"  ,"$5"  ,"$6"  ,"$7" 
,"$8" ,"$9" ,"$10" ,"$11" ,"$12" ,"$13" ,"$14" ,"$15"
,"$16" ,"$17" ,"$18" ,"$19" ,"$20" ,"$21" ,"$22" ,"$23"
,"$24" ,"$25" ,"$26" ,"$27" ,"$28" ,"$29" ,"$30" ,"$31"
};

/* Variables needed by the disassembly functions */
char **regs=algor_regs;
char **CP0regs=r3k_CP0regs;
isa_t isa=mips3;
int debug=0;

/*
 * Opcodes
 */
insn_st opcodes[64] = {
  {NULL, TYPE_SPECIAL, mips1},	{NULL, TYPE_REGIMM, mips1},
  {"j", TYPE_JUMP, mips1},	{"jal", TYPE_JUMP, mips1},
  {"beq", TYPE_BRA, mips1},	{"bne", TYPE_BRA, mips1},
  {"blez", TYPE_BRA, mips1},	{"bgtz", TYPE_BRA, mips1},
  {"addi", TYPE_IMM, mips1},	{"addiu", TYPE_IMM, mips1},
  {"slti", TYPE_IMMU, mips1},	{"sltiu", TYPE_IMMU, mips1},
  {"andi", TYPE_IMMU, mips1},	{"ori", TYPE_IMMU, mips1},
  {"xori", TYPE_IMMU, mips1},	{"lui", TYPE_IMMU, mips1},
  {"cop0", TYPE_CP, mips1},	{"cop1", TYPE_CP, mips1},
  {"cop2", TYPE_CP, mips1},	{"cop3", TYPE_CP, mips1},
  {"beql", TYPE_BRA, mips2},	{"bnel", TYPE_BRA, mips2},
  {"blezl", TYPE_BRA, mips2},	{"bgtzl", TYPE_BRA, mips2},
  {"daddi", TYPE_IMM, mips3},	{"daddiu", TYPE_IMM, mips3},
  {"ldl", TYPE_LDST, mips2},	{"ldr", TYPE_LDST, mips3},
  {NULL, TYPE_RES, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips1},	{NULL, TYPE_RES, mips0},
  {"lb", TYPE_LDST, mips1},	{"lh", TYPE_LDST, mips1},
  {"lwl", TYPE_LDST, mips1},	{"lw", TYPE_LDST, mips1},
  {"lbu", TYPE_LDST, mips1},	{"lhu", TYPE_LDST, mips1},
  {"lwr", TYPE_LDST, mips1},	{"lwu", TYPE_LDST, mips3},
  {"sb", TYPE_LDST, mips1},	{"sh", TYPE_LDST, mips1},
  {"swl", TYPE_LDST, mips1},	{"sw", TYPE_LDST, mips1},
  {"sdl", TYPE_LDST, mips3},	{"sdr", TYPE_LDST, mips3},
  {"swr", TYPE_LDST, mips1},	{"cache", TYPE_CACHE, mips2}, /* FIXME: is CACHE mips-2 or mips-3? */
  {NULL, TYPE_ISA, mips1},	{"lwc1", TYPE_LDST, mips1}, /* lwc0(mips1)->ll(mips2). Both are LDST */
  {"lwc2", TYPE_LDST, mips1},	{"lwc3", TYPE_LDST, mips1},
  {"lld", TYPE_LDST, mips3},	{"ldc1", TYPE_LDST, mips3},
  {"ldc2", TYPE_LDST, mips3},	{"ld", TYPE_LDST, mips3},
  {NULL, TYPE_ISA, mips1},	{"swc1", TYPE_LDST, mips1}, /* swc0(mips1)->sc(mips2). Both are LDST */
  {"swc2", TYPE_LDST, mips1},	{"swc3", TYPE_LDST, mips1},
  {"scd", TYPE_LDST, mips3},	{"sdc1", TYPE_LDST, mips3},
  {"sdc2", TYPE_LDST, mips3},	{"sd", TYPE_LDST, mips3}
};

insn_st special[64] = {
  {"sll", TYPE_SHIFT, mips1},	{NULL, TYPE_RES, mips0},
  {"srl", TYPE_SHIFT, mips1},	{"sra", TYPE_SHIFT, mips1},
  {"sllv", TYPE_REG, mips1},	{NULL, TYPE_RES, mips0},
  {"srlv", TYPE_REG, mips1},	{"srav", TYPE_REG, mips1},
  {"jr", TYPE_JR, mips1},	{"jalr", TYPE_JR, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"syscall", TYPE_SYSCALL, mips1},	{"break", TYPE_SYSCALL, mips1},
  {NULL, TYPE_RES, mips0},	{"sync", TYPE_NOP, mips2},
  {"mfhi", TYPE_HILO, mips1},	{"mthi", TYPE_HILO, mips1},
  {"mflo", TYPE_HILO, mips1},	{"mtlo", TYPE_HILO, mips1},
  {"dsllv", TYPE_REG, mips3},	{NULL, TYPE_RES, mips0},
  {"dsrlv", TYPE_REG, mips2},	{"dsrav", TYPE_REG, mips3},
  {"mult", TYPE_MOVE, mips1},	{"multu", TYPE_MOVE, mips1},
  {"div", TYPE_MOVE, mips1},	{"divu", TYPE_MOVE, mips1},
  {"dmult", TYPE_MOVE, mips3},	{"dmultu", TYPE_MOVE, mips3},
  {"ddiv", TYPE_MOVE, mips3},	{"ddivu", TYPE_MOVE, mips3},
  {"add", TYPE_REG, mips1},	{"addu", TYPE_REG, mips1},
  {"sub", TYPE_REG, mips1},	{"subu", TYPE_REG, mips1},
  {"and", TYPE_REG, mips1},	{"or", TYPE_REG, mips1},
  {"xor", TYPE_REG, mips1},	{"nor", TYPE_REG, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"slt", TYPE_REG, mips1},	{"sltu", TYPE_REG, mips1},
  {"dadd", TYPE_REG, mips3},	{"daddu", TYPE_REG, mips3},
  {"dsub", TYPE_REG, mips3},	{"dsubu", TYPE_REG, mips3},
  {"tge", TYPE_MOVE, mips2},	{"tgeu", TYPE_MOVE, mips2},
  {"tlt", TYPE_MOVE, mips2},	{"tltu", TYPE_MOVE, mips2},
  {"teq", TYPE_MOVE, mips2},	{NULL, TYPE_RES, mips0},
  {"tne", TYPE_MOVE, mips2},	{NULL, TYPE_RES, mips0},
  {"dsll", TYPE_SHIFT, mips3},	{NULL, TYPE_RES, mips0},
  {"dsrl", TYPE_SHIFT, mips3},	{"dsra", TYPE_SHIFT, mips3},
  {"dsll32", TYPE_SHIFT, mips3},	{NULL, TYPE_RES, mips0},
  {"dsrl32", TYPE_SHIFT, mips3},	{"dsra32", TYPE_SHIFT, mips3}
};

insn_st regimm[32] = {
  {"bltz", TYPE_BRA, mips1},	{"bgez", TYPE_PRINT, mips1},
  {"bltzl", TYPE_BRA, mips2},	{"bgezl", TYPE_PRINT, mips2},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"tgei", TYPE_TRAPI, mips2},	{"tgeiu", TYPE_TRAPI, mips2},
  {"tlti", TYPE_TRAPI, mips2},	{"tltiu", TYPE_TRAPI, mips2},
  {"teqi", TYPE_TRAPI, mips2},	{NULL, TYPE_RES, mips0},
  {"tnei", TYPE_TRAPI, mips2},	{NULL, TYPE_RES, mips0},
  {"bltzal", TYPE_BRA, mips1},	{"bgezal", TYPE_PRINT, mips1},
  {"bltzall", TYPE_BRA, mips2},	{"bgezall", TYPE_PRINT, mips2},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
};

fixup_t fixup_insns[] = {
  {"lwc0", TYPE_NONE, TYPE_LDST, mips1, 0x1a},
  {"ll", TYPE_NONE, TYPE_LDST, mips2, 0x1a},
  {"ll", TYPE_NONE, TYPE_LDST, mips3, 0x1a},
  {"b", TYPE_REGIMM, TYPE_BRA, mips1, 0x01},
  {"bl", TYPE_REGIMM, TYPE_BRA, mips1, 0x3},
  {"bal", TYPE_REGIMM, TYPE_BRA, mips1, 0x11},
  {"ball", TYPE_REGIMM, TYPE_BRA, mips2, 0x13},
  {"rfe", TYPE_CP, TYPE_NOP, mips1, 0x10},
  {"eret", TYPE_CP, TYPE_NOP, mips2, 0x10},
  {"eret", TYPE_CP, TYPE_NOP, mips3, 0x10},
  {"b", TYPE_NONE, TYPE_BRA, mips1, 0x04},
  {"li", TYPE_NONE, TYPE_IMM, mips1, 0x09},
  {NULL, TYPE_NONE, TYPE_NONE, mips0, 0x00} /* end marker */
};

insn_st COPzrs[16] = {
  {"mf", TYPE_CP, mips1},	{"dmf", TYPE_CP, mips3},
  {"cf", TYPE_CP, mips1},	{NULL, TYPE_INV, mips0},
  {"mt", TYPE_CP, mips1},	{"dmt", TYPE_CP, mips3},
  {"ct", TYPE_CP, mips1},	{NULL, TYPE_INV, mips0},
  {"bc", TYPE_CP, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
  /* The second half of this table is CoProc specific */
};
  
insn_st COPzrt[32] = {
  {"bcf", TYPE_CP, mips1},	{"bct", TYPE_CP, mips1},
  {"bcfl", TYPE_CP, mips2},	{"bctl", TYPE_CP, mips2},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0},
  {NULL, TYPE_INV, mips0},	{NULL, TYPE_INV, mips0}
};
  
insn_st CP0_opcode[64] = {
  {NULL, TYPE_RES, mips0},	{"tlbr", TYPE_NOP, mips1},
  {"tlbwi", TYPE_NOP, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"tlbwr", TYPE_NOP, mips1},	{NULL, TYPE_RES, mips0},
  {"tlbp", TYPE_NOP, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"BUG", TYPE_ISA, mips1},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"wait", TYPE_NOP, mips3},	{NULL, TYPE_RES, mips0}, /* wait is a R4600 specific insn */
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
};

insn_st CP1_func[64] = {
  {"add", TYPE_REG, mips1},	{"sub", TYPE_REG, mips1},
  {"mul", TYPE_REG, mips1},	{"div", TYPE_REG, mips1},
  {"sqrt", TYPE_MOVE, mips1},	{"abs", TYPE_MOVE, mips1},
  {"mov", TYPE_MOVE, mips1},	{"neg", TYPE_MOVE, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"round.w", TYPE_MOVE,mips1},	{"trunc.w", TYPE_MOVE, mips1},
  {"ceil.w", TYPE_MOVE, mips1},	{"floor.w", TYPE_MOVE, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"cvt.s", TYPE_MOVE, mips1},	{"cvt.d", TYPE_MOVE, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"cvt.w", TYPE_MOVE, mips1},	{"cvt.l", TYPE_MOVE, mips3}, /* cvt.l is possibly a mips1 insn */
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"c.f", TYPE_BRA, mips1},	{"c.un", TYPE_BRA, mips1},
  {"c.eq", TYPE_BRA, mips1},	{"c.ueq", TYPE_BRA, mips1},
  {"c.olt", TYPE_BRA, mips1},	{"c.ult", TYPE_BRA, mips1},
  {"c.ole", TYPE_BRA, mips1},	{"c.ule", TYPE_BRA, mips1},
  {"c.sf", TYPE_BRA, mips1},	{"c.ngle", TYPE_BRA, mips1},
  {"c.seq", TYPE_BRA, mips1},	{"c.ngl", TYPE_BRA, mips1},
  {"c.lt", TYPE_BRA, mips1},	{"c.nge", TYPE_BRA, mips1},
  {"c.le", TYPE_BRA, mips1},	{"c.ngt", TYPE_BRA, mips1}
};

insn_st CP1_fmt[16] = {
  {"s", TYPE_NOP, mips1},	{"d", TYPE_NOP, mips1},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {"w", TYPE_NOP, mips1},	{"l", TYPE_NOP, mips3},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0},
  {NULL, TYPE_RES, mips0},	{NULL, TYPE_RES, mips0}
};

char *r3k_CP0regs[32] = {
  "index",	"random",	"entrylo",	"3",
  "context",	"5",		"6",		"7",
  "bvaddr",	"9",		"entryhi",	"11",
  "sr",		"cause",	"epc",		"prid",
  "16",		"17",		"18",		"19",
  "20",		"21",		"22",		"23",
  "24",		"25",		"26",		"27",
  "28",		"29",		"30",		"31"
};

char *r4k_CP0regs[32] = {
  "index",	"random",	"entrylo",	"3",
  "context",	"pagemask",	"wired",	"error",
  "bvaddr",	"count",	"entryhi",	"compare",
  "sr",		"cause",	"epc",		"prid",
  "config",	"lladdr",	"watchlo",	"watchhi",
  "20",		"21",		"22",		"23",
  "24",		"25",		"ecc",		"cacheerr",
  "taglo",	"taghi",	"errepc",	"31"
};














inline int
sign_extend(int imm)
{
#if 0
  return (imm & (1 << 15)) ? (~0<<16) | imm : imm;
#else
  /* I guess this aught to generate optimum code on all platforms */
  return (int) ((short) imm);
#endif
}

void
pr_bug(char *string, unsigned insn)
{
  sprintf(string,".word    0x%08x    # ***BUG***",insn);
  return;
}

void
pr_print(char *string, insn_st **class, unsigned insn)
{
  /* stub function */
  return;
}

void
pr_isa(char *string, insn_st **class, unsigned insn)
{
  /* stub function */
  return;
}

void
pr_inv(char *string, unsigned insn)
{
  sprintf(string,".word    0x%08x    # invalid instruction",insn);
  return;
}

void
pr_res(char *string, unsigned insn)
{
  sprintf(string,".word    0x%08x    # Reserved instruction (causes "
	  "exception)",insn);
  return;
}

void
pr_none(char *string, unsigned insn)
{
  /* This shouldn't happen */
  sprintf(string,".word    0x%08x    # Whoops! - internal error.",insn);
  return;
}

void
pr_nop(char *string, insn_st *insn)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s",insn->name);
  } else {
    sprintf(string,"%s    # (mips-%i insn)",insn->name,insn->isa);
  }
  return;
}  


#if 0
void
pr_special(void)
{
  /* TYPE_SPECIAL isn't a printed type as such; TYPE_SPECIAL means
   * that we should look in the [5..0] field for the sub-opcode
   */
  return;
}
#endif

void
pr_regimm(char *string, unsigned insn, unsigned addr)
{
  unsigned r1, r2;
  unsigned imm16;

  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  imm16 = FIELD(insn,0,16);

  switch(regimm[r2].type) {
  case TYPE_BRA:
    pr_bra(string,&regimm[r2],r1,addr,imm16);
    break;
  case TYPE_TRAPI:
    pr_trapi(string,&regimm[r2],r1,imm16);
    break;
  case TYPE_RES:
    pr_res(string,insn);
    break;
  case TYPE_INV:
    pr_inv(string,insn);
    break;
  case TYPE_PRINT:
    /* Ugh, right. Need to do something intelligent here. */
    pr_bra(string,&regimm[r2],r1,addr,imm16); /* FIXME */
    break;
  default:
    pr_bug(string,insn);
    break;
  } /* switch(regimm[r2].type) */
}

void
pr_immu(char *string, insn_st *insn, int r1, int r2, unsigned imm)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s,%s,0x%04x",insn->name,regs[r2],regs[r1],imm);
  } else {
    sprintf(string,"%s    %s,%s,0x%04x    # (mips-%i insn)",insn->name,
	    regs[r2],regs[r1],imm,insn->isa); 
  }
  return;
}  

void
pr_reg(char *string, insn_st *insn, int r1, int r2, int r3)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s,%s,%s", insn->name,
	    regs[r3],regs[r1],regs[r2]);
  } else {
    sprintf(string,"%s    %s,%s,%s    # (mips-%i insn)", insn->name,
	    regs[r3],regs[r1],regs[r2],insn->isa);
  }
  return;
}

void
pr_jump(char *string, insn_st *insn, unsigned addr, unsigned offset)
{
  unsigned long target = (addr & (~0 << 28)) | (offset << 2);
  if(isa>=insn->isa) {
    sprintf(string,"%s    %08lx",insn->name,target);
  } else {
    sprintf(string,"%s    %08lx    # (mips-%i insn)",insn->name,target,insn->isa);
  }
  return;
}

void
pr_imm(char *string, insn_st *insn, int r1, int r2, int imm)
{
  imm = sign_extend(imm);
  if(isa>=insn->isa) {
    if(imm >= 0 ) {
      sprintf(string,"%s    %s,%s,0x%04x",insn->name,
	      regs[r2],regs[r1],imm);
    } else {
      sprintf(string,"%s    %s,%s,%i",insn->name,
	      regs[r2],regs[r1],imm);
    }
  } else {
    if(imm >= 0 ) {
      sprintf(string,"%s    %s,%s,0x%04x    # (mips-%i insn)",insn->name,
	      regs[r2],regs[r1],imm,insn->isa);
    } else {
      sprintf(string,"%s    %s,%s,%i    # (mips-%i insn)",insn->name,
	      regs[r2],regs[r1],imm,insn->isa);
    }
  }
  return;
}

void
pr_bra(char *string, insn_st *insn, int r1, unsigned addr,
       unsigned imm)
{
  unsigned long target = addr + 4 + (imm << 2);
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s,%08lx",insn->name,regs[r1],target);
  } else {
    sprintf(string,"%s    %s,%08lx    # (mips-%i insn)",insn->name,regs[r1],
	    target,insn->isa);
  }
  return;
}

void
pr_ldst(char *string, insn_st *insn, int r1, int r2, int offset)
{
  offset = sign_extend(offset);
  if(isa>=insn->isa) {
    if(offset>=0) {
      sprintf(string,"%s    %s,0x%x(%s)",insn->name,regs[r2],offset,regs[r1]);
    } else {
      sprintf(string,"%s    %s,%i(%s)",insn->name,regs[r2],offset,regs[r1]);
    }
  } else {
    if(offset>=0) {
      sprintf(string,"%s    %s,0x%x(%s)    # (mips-%i insn)",insn->name,
	      regs[r2],offset,regs[r1],insn->isa);
    } else {
      sprintf(string,"%s    %s,%i(%s)    # (mips-%i insn)",insn->name,regs[r2],
	      offset,regs[r1],insn->isa);
    }
  }
  return;
}

void
pr_shift(char *string, insn_st *insn, int r2, int r3, int shift)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s,%s,0x%02x",insn->name,regs[r3],regs[r2],shift);
  } else {
    sprintf(string,"%s    %s,%s,0x%02x    # (mips-%i insn)",insn->name,
	    regs[r3],regs[r2],shift,insn->isa);
  }
  return;
}

void
pr_jr(char *string, insn_st *insn, int reg)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s",insn->name,regs[reg]);
  } else {
    sprintf(string,"%s    %s    # (mips-%i insn)",insn->name,regs[reg],insn->isa);
  }
  return;
}

void
pr_syscall(char *string, insn_st *insn, unsigned imm)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    0x%05x",insn->name,imm);
  } else {
    sprintf(string,"%s    0x%05x    # (mips-%i insn)",insn->name,imm,insn->isa);
  }
  return;
}

void
pr_hilo(char *string, insn_st *insn, int reg)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s",insn->name,regs[reg]);
  } else {
    sprintf(string,"%s    %s    # (mips-%i insn)",insn->name,regs[reg],insn->isa);
  }
  return;
}

void
pr_move(char *string, insn_st *insn, int r1, int r2)
{
  if(isa>=insn->isa) {
    sprintf(string,"%s    %s,%s",insn->name,regs[r1],regs[r2]);
  } else {
    sprintf(string,"%s    %s,%s    # (mips-%i insn)",insn->name,regs[r1],
	    regs[r2],insn->isa);
  }
  return;
}

void
pr_cp(void)
{
  /* I guess the CPn stuff need a /bit/ more handling :-) */
  return;
}

void
pr_cache(char *string, insn_st *insn, int r1, int r2, int imm)
{
  imm = sign_extend(imm);
  if(isa>=insn->isa) {
    if(imm < 0) {
      sprintf(string,"%s    0x%x,%i(%s)",insn->name,r2,imm,regs[r1]);
    } else {
      sprintf(string,"%s    0x%x,0x%x(%s)",insn->name,r2,imm,regs[r1]);
    }
  } else {
    if(imm < 0) {
      sprintf(string,"%s    0x%x,%i(%s)    # (mips-%i insn)",insn->name,r2,imm,
	      regs[r1],insn->isa);
    } else {
      sprintf(string,"%s    0x%x,0x%x(%s)     (mips-%i insn)",insn->name,r2,imm,
	      regs[r1],insn->isa);
    }
  }
  return;
}

void
pr_trapi(char *string, insn_st *insn, int reg, int imm)
{
  imm = sign_extend(imm);
  if(isa>=insn->isa) {
    if(imm>=0) {
      sprintf(string,"%s    %s,0x%04x",insn->name,regs[reg],imm);
    } else {
      sprintf(string,"%s    %s,%i",insn->name,regs[reg],imm);
    }
  } else {
    if(imm>=0) {
      sprintf(string,"%s    %s,0x%04x    # (mips-%i insn)",insn->name,
	      regs[reg],imm,insn->isa);
    } else {
      sprintf(string,"%s    %s,%i    # (mips-%i insn)",insn->name,regs[reg],
	      imm,insn->isa);
    }
  }
  return;
}

void
print_special(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned r1, r2, r3;
  unsigned shift;
  unsigned imm16;
  unsigned imm26;

  signed int imm;

  /*
   * Well, the *real* opcode field lies at [31..26], but we look at
   * the sub-opcode here (main opcode is "special").
   */
  opcode = FIELD(insn,0,6);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  imm16 = FIELD(insn,0,16);
  imm26 = FIELD(insn,0,26);
  shift = FIELD(insn,6,5);

  switch(special[opcode].type) {
  case TYPE_REG:
    pr_reg(string,&special[opcode],r1,r2,r3);
    break;
  case TYPE_IMM:
    /* convert imm16 into a "true" signed int */
    imm = sign_extend(imm16);
    if(imm >= 0 ) {
      sprintf(string,"%s    %s,%s,0x%04x",special[opcode].name,
	      regs[r2],regs[r1],imm);
    } else {
      sprintf(string,"%s    %s,%s,%i",special[opcode].name,
	      regs[r2],regs[r1],imm);
    }
    break;
  case TYPE_SHIFT:
    sprintf(string,"%s    %s,%s,0x%02x",special[opcode].name,regs[r3],
	    regs[r2],shift);
    break;
  case TYPE_JR:
    sprintf(string,"%s    %s",special[opcode].name,regs[r1]);
    break;
  case TYPE_SYSCALL:
    /* extract "code" field */
    imm = FIELD(insn,6,20);
    sprintf(string,"%s    0x%05x",special[opcode].name,imm);
    break;
  case TYPE_NOP:
    sprintf(string,"%s",special[opcode].name);
    break;
  case TYPE_HILO:
    sprintf(string,"%s    %s",special[opcode].name,regs[r2]);
    break;
  case TYPE_MOVE:
    sprintf(string,"%s    %s,%s",opcodes[opcode].name,regs[r1],regs[r2]);
    break;
  case TYPE_NONE:
    /* illegal insn */
    sprintf(string,".word    0x%08x    # illegal or reserved "
	    "instruction",insn);
    if(debug) {
      sprintf(string+strlen(string),"\n    # in print_special():"
	      "switch(opcodes[].type)");
    }
    break;
  default:
    /* Internal error -- cannot happen! */
    sprintf(string,"Oops!    internal error in print_special().");
    break;
  }
}


/*
 * parse a COPz insn
 */
void
print_cp(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned cp;
  unsigned r1, r2, r3, r4;

  /*
   * Well, the true opcode field lies at [31..26], but we have
   * already decode that, and now turns to the sub-opcode at
   * [5..0] (main opcode is "COPz").
   */
  opcode = FIELD(insn,0,6);
  cp = FIELD(insn,26,2);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  r4 = FIELD(insn,6,5);

  switch(cp) {
  case 0: /* System co-proc */
    /* First decide if it is a tlbxx/eret/wait type insn */
    if(FIELD(insn,25,1)) {
      /* Yup, it is.  Look up the insn in the CP0_opcode[] */
      switch(CP0_opcode[opcode].type) {
      case TYPE_NOP:
	sprintf(string,"%s",CP0_opcode[opcode].name);
	break;
      case TYPE_NONE:
	sprintf(string,".word    0x%08x    # illegal or reserved "
		"instruction",insn);
	if(debug) {
	  sprintf(string+strlen(string),"\n    # print_cp():CPzrs(%i):"
		  "TYPE_NONE",cp);
	}
	break;
      default: /* This cannot happen */
	sprintf(string,"print_cp(): This cannot happen! "
		"opcode = 0x%02x",opcode);
	break;
      } /* switch(CP0_opcode[opcode].type) */
    } else { /* if(!FIELD(insn,25,1)) */
      switch(COPzrs[r1].type) {
      case TYPE_CP:
	sprintf(string,"%sc0    %s,%s",COPzrs[r1].name,
		regs[r2],CP0regs[r3]);
	break;
      default:
	sprintf(string,".word    0x%08x    # illegal or reserved "
		"instruction",insn);
	if(debug) {
	  sprintf(string+strlen(string),"illegal    # print_cp():CPzrs(%i):"
		  "default",cp);
	}
	break;
      } /* switch(COPzrs) */
    } /* if(FIELD(insn,25,1)) */
    break;
    /* END OF CP0 PROCESSING */
  case 1: /* Floating Point co-proc */
    if(opcode == 0x00) {
      /* "standard" move from/to CP1 insns */
      switch(COPzrs[r1].type) {
      case TYPE_CP:
	sprintf(string,"%sc1    %s,f%i",COPzrs[r1].name,
		regs[r2],r3);
	break;
      case TYPE_NOP:
	/* Naughty, naughty:  --FIXME!!!! */ 
	if(r1 == 16 || r1 == 17) {
	  sprintf(string,"%s.%s    f%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r3,r2);
	} else {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill    f%i,f%i",CP1_func[opcode].name,
		  r3,r2);
	}
	break;
      case TYPE_NONE:
	sprintf(string,".word    0x%08x",insn);
	break;
      default: /* This cannot happen */
	sprintf(string,"print_cp(): This cannot happen! "
		"opcode = 0x%02x",opcode);
	break;
      } /* switch(COPzrs[r1].type) */
    } else {
      /* "real" FPU insns */
      switch(CP1_func[opcode].type) {
      case TYPE_REG:
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill    f%i,f%i,f%i",
		  CP1_func[opcode].name,r4,r3,r2);
	} else {
	  sprintf(string,"%s.%s    f%i,f%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r4,r3,r2);
	}
	break;
      case TYPE_MOVE:
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_NONE) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill    f%i,f%i",CP1_func[opcode].name,
		  r4,r3);
	} else {
	  sprintf(string,"%s.%s    f%i,f%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r4,r3,r2);
	}
	break;
      case TYPE_BRA:
	/* This is *compares* not branches(!) */
	/* C.cond.fmt fA,fB */
	if(r1<16 || CP1_fmt[r1-16].type == TYPE_RES) {
	  /* illegal format specifier */
	  sprintf(string,"%s.ill    f%i,f%i",CP1_func[opcode].name,
		  r3,r2);
	} else {
	  sprintf(string,"%s.%s    f%i,f%i",CP1_func[opcode].name,
		  CP1_fmt[r1].name,r3,r2);
	}
	break;
      case TYPE_NONE:
	sprintf(string,".word    0x%08x    # illegal or reserved "
		"instruction",insn);
	if(debug) {
	  sprintf(string+strlen(string),"\n    # in print_cp():"
		  "CP1_func:TYPE_NONE");
	}
	break;
      default:
	sprintf(string,"    # Oops! internal error in print_cp():CP1_func)");
	break;
      } /* switch(CP1_func[opcode].type) */
    } /* !if(opcode == 0x00) */
    break;
    /* END OF CP1 PROCESSING */
  case 2: /* ??? co-proc */
    sprintf(string,".word    0x%08x    # cp%i: giving up!",insn,cp);
    break;
  case 3: /* invalid?? co-proc */
    sprintf(string,".word    0x%08x    # cp%i: illegal coproc",insn,cp);
    break;
  default: /* cannot happen */
    sprintf(string,"print_cp(): This cannot happen. cp = %i",cp);
    break;
  }
  return;
}

/*
 * examine the machine instruction `insn' and format it into `string',
 * under the assumption that it resides at `addr'.
 */
void
mips_disassemble(char *string, unsigned insn, unsigned addr)
{
  unsigned opcode;
  unsigned r1, r2, r3;
  unsigned shift;
  unsigned imm16;
  unsigned imm26;
  unsigned offset;
  signed int imm;
  unsigned long target;

  opcode = FIELD(insn,26,6);
  r1 = FIELD(insn,21,5);
  r2 = FIELD(insn,16,5);
  r3 = FIELD(insn,11,5);
  imm16 = FIELD(insn,0,16);
  imm26 = FIELD(insn,0,26);
  offset = FIELD(insn,0,6);
  shift = FIELD(insn,6,6);

  /*
   * Handle all "special cases" first
   */
  switch(opcodes[opcode].type) {
  case TYPE_PRINT:
    pr_bug(string,insn);
    break;
  case TYPE_SPECIAL:
    /* "sll zero,zero,0" => "nop" */
    if(insn == 0x00) {
      sprintf(string,"nop");
      return;
    }
    /* map "or rA,zero,rB" => "move rA,rB" */
    if(offset == 0x25 && r2 == 0x00) {
      sprintf(string,"move    %s,%s",regs[r3],regs[r1]);
      return;
    }
    break;
  case TYPE_REGIMM:
#if 0 /* comment this out so I can see the full ghastness of it */
    if(regimm[r2].type == TYPE_PRINT && r1 == 0) {
      fixup_t *tmp_insn;
      int flag=0;
      target = (unsigned long) (addr + 4 + (sign_extend(imm16)<<2));
      i=0;
      do {
	tmp_insn = fixup_insns[i];
	if(tmp_insn.type==TYPE_REGIMM && tmp_insn.opcode==r2) {
	  flag=1;
	  /* match */
	  if(tmp_insn.decode==TYPE_BRA) {
	    sprintf(string,"%s    %08lx",tmp_insn.name,target);
	  } else {
	    pr_bug(insn);
	  }
	}
	i++;
      } while(fixup_insn[i].name!=NULL && !flag);
      if!(flag) pr_bug(insn);
    } /* if(regimm[r2].type == TYPE_BRA && r1 == 0) */
#endif
    break;
  case TYPE_IMMU:
    /* "lui rA,zero,imm" => "lui rA,imm" */
    if(opcode == 0x0f && r1 == 0x00) {
      sprintf(string,"lui    %s,0x%04x",regs[r2],imm16);
      return;
    }
    break;
  case TYPE_IMM:
    /* "addiu rA,zero,imm" => "li rA,imm" */
    if(opcode == 0x09 && r1 == 0x00) {
      imm = sign_extend(imm16);
      if(imm>=0) {
	sprintf(string,"li    %s,0x%04x",regs[r2],imm);
      } else {
	sprintf(string,"li    %s,%i",regs[r2],imm);
      }
      return;
    }
    break;
  case TYPE_BRA:
    /* "beq zero,zero,target" => "b target" */
    if(r1 == 0 && r2 == 0) {
      /* shift imm16 two bits left and signed extend it */
      target = (unsigned long)(addr + 4 + (sign_extend(imm16) << 2));
      sprintf(string,"b    %08lx",target);
      return;
    }
    /* "bCC r1,zero,target" => "bCCz r1,target" */
    if(r1 == 0 || r2 == 0) {
      /* compute target address */
      target = (unsigned long)(addr + 4 + (sign_extend(imm16)<<2));
      r1 += r2; /* Nice hack which cover both cases graciously */
      sprintf(string,"%sz    %s,%08lx",opcodes[opcode].name,
	      regs[r1],target);
      return;
    }
    break;
  case TYPE_CP:
    if(COPzrs[r1].type == TYPE_CP && FIELD(insn,26,2) == 0
       && (r1 == 0x02 || r1 == 0x06)) {
      /* CFCz and CTCz are illegal for COP0 */
      sprintf(string,".word    =x%08x    # illegal or reserved "
	      "instruction",insn);
      return;
    }
    break;
  default:
    /* It is perfectly valid to take the default action. This just means that
     * the insn will be handled by the "normal case" code below */
    break;
  } /* Handle "special cases" */

  /*
   * Now handle the normal cases
   */
  switch(opcodes[opcode].type) {
  case TYPE_SPECIAL:
    print_special(string,insn,addr);
    break;
  case TYPE_REGIMM:
    pr_regimm(string,insn,addr);
    break;
  case TYPE_IMMU:
    sprintf(string,"%s    %s,%s,0x%04x",opcodes[opcode].name,regs[r2],
	    regs[r1],imm16);
    break;
  case TYPE_JUMP:
    target = (addr & (~0 << 28)) | (imm26 << 2);
    sprintf(string,"%s    %08lx",opcodes[opcode].name,target);
    break;
  case TYPE_IMM:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    if(imm >= 0) {
      sprintf(string,"%s    %s,%s,0x%04x",opcodes[opcode].name,regs[r2],
	      regs[r1],imm);
    } else {
      sprintf(string,"%s    %s,%s,%i",opcodes[opcode].name,regs[r2],regs[r1],
	      imm);
    }
    break;
  case TYPE_BRA:
    /* shift imm16 two bits left and signed extend it */
    imm = (imm16 & (1 << 15)) ? -(1<<18) + (imm16 << 2) : imm16 << 2;
    /* compute target address */
    target = (unsigned long)(addr + 4 + imm);
    sprintf(string,"%s    %s,%s,%08lx",opcodes[opcode].name,regs[r1],regs[r2],
	    target);
    break;
  case TYPE_LDST:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    sprintf(string,"%s    %s,0x%x(%s)",opcodes[opcode].name,regs[r2],imm,
	    regs[r1]);
    break;
  case TYPE_JR:
    sprintf(string,"%s    %s",opcodes[opcode].name,regs[r1]);
    break;
  case TYPE_CP:
    print_cp(string,insn,addr);
    break;
  case TYPE_CACHE:
    imm = (imm16 & (1 << 15)) ? -(1 << 16) + imm16 : imm16;
    if(imm < 0) {
      sprintf(string,"%s    0x%x,%i(%s)",opcodes[opcode].name,r2,imm,regs[r1]);
    } else {
      sprintf(string,"%s    0x%x,0x%x(%s)",opcodes[opcode].name,r2,imm,
	      regs[r1]);
    }
    break;
  case TYPE_NONE:
    sprintf(string,".word    0x%08x    # illegal or reserved "
	    "instruction",insn);
    if(debug) {
      sprintf(string+strlen(string),"\n    # in disassemble():Normal "
	      "cases:TYPE_NONE");
    }
    break;
  case TYPE_NOP:
  case TYPE_REG:
  case TYPE_SHIFT:
  case TYPE_SYSCALL:
  case TYPE_HILO:
  case TYPE_MOVE:
    /* These can only happen as a sub-opcode under SPECIAL */
  default:
    /* illegal insn or something wrong */
    sprintf(string,"Oops!    internal error in disassemble(): "
	    "opcode type = %03x",opcodes[opcode].type );
    break;
  }
}

/*
 * convert a big/little endian word into a little/big endian word
 */
inline unsigned long
nuxi (const unsigned long value)
{
  return ((value >> 24) & 0x000000ffUL)
       | ((value >>  8) & 0x0000ff00UL)
       | ((value <<  8) & 0x00ff0000UL)
       | ((value << 24) & 0xff000000UL);
}
